from pathlib import Path
from typing import Dict, List

import gemmi as gemmi

from structurefinder.searcher.fileparser import CifFile


def export_to_cif_file(cif: Dict, filename: str):
    doc = cif_data_to_document(cif)
    write_cif_to_disk(doc, filename)


def cif_data_to_document(cif: Dict) -> gemmi.cif.Document:
    doc: gemmi.cif.Document = gemmi.cif.Document()
    title = 'exported_from_structurefinder_{}'.format(cif['data'])
    block: gemmi.cif.Block = doc.add_new_block(title)
    for key, value in cif.items():
        if key == '_loop':
            add_loop_to_block(block, value)
            continue
        if not key.startswith('_') or key == '_':
            continue
        if key == '_space_group_symop_operation_xyz':
            xyzloop = []
            for val in value.splitlines(keepends=False):
                val = replace_float_values(val)
                xyzloop.append({'_space_group_symop_operation_xyz': val})
            add_loop_to_block(block, xyzloop)
            # continue to prevent overriding loop by _space_group_symop_operation_xyz key-value pair.
            continue
        str_value = str(value)
        if not str_value:
            str_value = '?'
        block.set_pair(key, gemmi.cif.quote(str_value))
    return doc


def write_cif_to_disk(doc, filename):
    doc.write_file(filename, style=gemmi.cif.Style.Indent35)


def replace_float_values(val: str) -> str:
    val = val.replace('0.75', '3/4')
    val = val.replace('0.5', '1/2')
    val = val.replace('0.33', '1/3')
    val = val.replace('0.25', '1/4')
    val = val.replace('0.125', '1/6')
    return val


def add_loop_to_block(block: 'gemmi.cif.Block', value: List[Dict]) -> None:
    loop: Dict
    new_loop = None
    current_loop = None
    for loop in value:
        row = gemmi.cif.quote_list(list(loop.values()))
        if not current_loop:
            new_loop = block.init_loop('', list(loop.keys()))
            new_loop.add_row(row)
            current_loop = loop.keys()
        else:
            if loop.keys() == current_loop:
                new_loop.add_row(row)
            else:
                current_loop = None
    return None


if __name__ == '__main__':
    cif = CifFile()
    for num, file in enumerate(Path('./test-data').rglob('*.cif')):
        doc = gemmi.cif.Document()
        doc.source = file
        try:
            doc.parse_file(file)
        except ValueError:
            continue
        cifok = cif.parsefile(doc)
        export_to_cif_file(cif.cif_data, 'test{}.cif'.format(num))
        # if num == 2:
        #    break



import os
import re
import time
from typing import Optional, Union

import gemmi
from PyQt5 import QtCore

from structurefinder.searcher.database_handler import StructureTable
from structurefinder.searcher.filecrawler import excluded_names, filewalker_walk, fill_db_with_cif_data, MyZipReader, \
    MyTarReader, fill_db_with_res_data
from structurefinder.searcher.fileparser import CifFile
from structurefinder.shelxfile.shelx import ShelXFile

DEBUG = False


class Worker(QtCore.QObject):
    finished = QtCore.pyqtSignal(str)
    progress = QtCore.pyqtSignal(int)
    number_of_files = QtCore.pyqtSignal(int)

    def __init__(self, searchpath: str, add_res_files: bool, add_cif_files: bool, lastid: int,
                 structures: StructureTable, excludes: Optional[list] = None, standalone: Optional[bool] = False):
        super().__init__()
        self.stop = False
        self.searchpath = searchpath
        self.add_res_files = add_res_files
        self.add_cif_files = add_cif_files
        self.structures = structures
        self.lastid = lastid
        self.excludes = excludes
        self.files_indexed = 0
        if standalone:
            self.files_indexed = self.index_files()

    def index_files(self):
        return self.put_files_in_db(searchpath=self.searchpath, structures=self.structures,
                                    fillres=self.add_res_files, excludes=None,
                                    fillcif=self.add_cif_files, lastid=self.lastid)

    def put_files_in_db(self, searchpath: str = '', excludes: Union[list, None] = None, lastid: int = 1,
                        structures=None, fillcif=True, fillres=True) -> int:
        """
        Imports files from a certain directory
        """
        if self.excludes:
            excluded_names.extend(excludes)
        if not searchpath:
            return 0
        if lastid <= 1:
            lastid = 1
        num = 1
        zipcifs = 0
        rescount = 0
        cifcount = 0
        time1 = time.perf_counter()
        patterns = ['*.cif', '*.zip', '*.tar.gz', '*.tar.bz2', '*.tgz', '*.res']
        filelist = filewalker_walk(str(searchpath), patterns)
        self.number_of_files.emit(len(filelist))
        options = {}
        filecount = 1
        for filenum, (filepth, name) in enumerate(filelist, start=1):
            if self.stop:
                structures.database.commit_db()
                self.finished.emit('Indexing aborted')
                return 0
            filecount = filenum
            fullpath = os.path.join(filepth, name)
            options['modification_time'] = time.strftime('%Y-%m-%d', time.gmtime(os.path.getmtime(fullpath)))
            options['file_size'] = int(os.stat(str(fullpath)).st_size)
            cif = CifFile(options=options)
            self.progress.emit(filecount)
            self.number_of_files.emit(filecount)
            # This is really ugly copy&pase code. TODO: refractor this:
            if name.endswith('.cif') and fillcif:
                #print(fullpath, '###')
                doc = gemmi.cif.Document()
                doc.source = fullpath
                try:
                    doc.parse_file(fullpath)
                except ValueError:
                    continue
                cifok = cif.parsefile(doc)
                if not cifok:
                    if DEBUG:
                        print(f"Could not parse (.cif): {fullpath}")
                    continue
                if cif:  # means cif object has data inside (cif could be parsed)
                    tst = None
                    try:
                        tst = fill_db_with_cif_data(cif, filename=name, path=filepth, structure_id=lastid,
                                                    structures=structures)
                    except Exception as err:
                        if DEBUG:
                            print(
                                str(err) + f"\nIndexing error in file {filepth}{os.path.sep}{name} - Id: {lastid}")
                            raise
                        continue
                    if not tst:
                        continue
                    cifcount += 1
                    num += 1
                    if lastid % 1000 == 0:
                        print(f'{num} files ...')
                        structures.database.commit_db()
                lastid += 1
                self.progress.emit(filecount)
                self.number_of_files.emit(filecount)
                continue
            if (name.endswith('.zip') or name.endswith('.tar.gz') or name.endswith('.tar.bz2')
                or name.endswith('.tgz')) and fillcif:
                if fullpath.endswith('.zip'):
                    # MyZipReader defines .cif ending:
                    z = MyZipReader(fullpath)
                else:
                    z = MyTarReader(fullpath)
                for zippedfile in z:  # the list of cif files in the zip file
                    if not zippedfile:
                        continue
                    # Important here to re-initialize empty cif dictionary:
                    cif = CifFile(options=options)
                    #print(zippedfile, 'z#i#p')
                    omit = False
                    for ex in excluded_names:  # remove excludes
                        if re.search(ex, z.cifpath, re.I):
                            omit = True
                    if omit:
                        continue
                    try:
                        cifok = cif.parsefile(zippedfile)
                        if not cifok:
                            if DEBUG:
                                print(f"Could not parse (zipped): {fullpath}")
                            continue
                    except IndexError:
                        continue
                    if cif:
                        tst = None
                        try:
                            tst = fill_db_with_cif_data(cif, filename=z.cifname, path=fullpath, structure_id=lastid,
                                                        structures=structures)
                        except Exception as err:
                            if DEBUG:
                                print(
                                    str(err) + f"\nIndexing error in file {filepth}{os.path.sep}{name} - Id: {lastid}")
                                raise
                            continue
                        if not tst:
                            if DEBUG:
                                print('cif file not added:', fullpath)
                            continue
                        zipcifs += 1
                        cifcount += 1
                        num += 1
                        if lastid % 1000 == 0:
                            print(f'{num} files ...')
                            structures.database.commit_db()
                        lastid += 1
                        self.progress.emit(filecount)
                        self.number_of_files.emit(filecount)
                continue
            if name.endswith('.res') and fillres:
                tst = None
                try:
                    res = ShelXFile(fullpath)
                except Exception as e:
                    if DEBUG:
                        print(e)
                        print(f"Could not parse (.res): {fullpath}")
                    continue
                if res:
                    tst = fill_db_with_res_data(res, filename=name, path=filepth, structure_id=lastid,
                                                structures=structures, options=options)
                if not tst:
                    if DEBUG:
                        print('res file not added:', fullpath)
                    continue
                num += 1
                rescount += 1
                if lastid % 1000 == 0:
                    print(f'{num} files ...')
                    structures.database.commit_db()
                lastid += 1
                self.progress.emit(filecount)
                self.number_of_files.emit(filecount)
        structures.database.commit_db()
        time2 = time.perf_counter()
        self.progress.emit(filecount)
        m, s = divmod(time2 - time1, 60)
        h, m = divmod(m, 60)
        tmessage = f'Added {num - 1} files ({cifcount} cif, {rescount} res) files ({zipcifs} in compressed files) ' \
                   f'to database in: {int(h):>2d} h, {int(m):>2d} m, {s:>3.2f} s'
        print(f'      {filecount} files considered.')
        print(tmessage)
        self.finished.emit(tmessage)
        return len(structures)



import argparse
import sys
import time
from argparse import Namespace
from pathlib import Path
from sqlite3 import DatabaseError

from structurefinder.misc.update_check import is_update_needed
from structurefinder.misc.version import VERSION
from structurefinder.pymatgen.core.lattice import Lattice
from structurefinder.searcher.database_handler import DatabaseRequest, StructureTable
from structurefinder.searcher.misc import vol_unitcell, regular_results_parameters
from structurefinder.searcher.worker import Worker

parser = argparse.ArgumentParser(description=f'Command line version {VERSION} of StructureFinder to collect .cif/.res files to a '
                                             'database.\n'
                                             'StructureFinder will search for cif files in the given directory(s) '
                                             'recursively.  (Either -c, -r or both options must be active!)')
parser.add_argument("-d",
                    dest="dir",
                    metavar='"directory"',
                    type=str,
                    action='append',
                    help='Directory(s) where cif files are located.')
parser.add_argument("-e",
                    dest="ex",
                    metavar='"directory"',
                    type=str,
                    action='append',
                    help='Directory names to be excluded from the file search. Default is:\n'
                         '"ROOT", ".OLEX", "TMP", "TEMP", "Papierkorb", "Recycle.Bin" '
                         'Modifying -e option discards the default.')
parser.add_argument("-o",
                    dest="outfile",
                    metavar='"sqlite file name"',
                    type=str,
                    help='Name of the output database file. Default: "./structuredb.sqlite"\n'
                         'Also used for the commandline search (-f option).')
parser.add_argument("-c",
                    dest="fillcif",
                    default=False,
                    action='store_true',
                    help='Add .cif files (crystallographic information file) to the database.')
parser.add_argument("-r",
                    dest="fillres",
                    default=False,
                    action='store_true',
                    help='Add SHELX .res files to the database.')
parser.add_argument("--delete",
                    dest="delete",
                    default=False,
                    action='store_true',
                    help="Delete and do not append to previous database.")
parser.add_argument("-f",
                    dest='cell',
                    # nargs=6,
                    metavar='"unit cell"',
                    type=lambda s: [float(item) for item in s.split()],
                    help='Search for the specified unit cell. \n'
                         'The cell values have to be enclosed in brackets.'
                    )
parser.add_argument("-m",
                    dest='merge',
                    metavar='"sqlite file name"',
                    default=False,
                    type=str,
                    help="Merges a database file into the file of '-o' option."
                    )


def check_update():
    if is_update_needed(VERSION=VERSION):
        print('A new Version of StructureFinder is available at '
              'https://dkratzert.de/structurefinder.html')


def find_cell(args: Namespace):
    """
    Searches for unit cells by command line parameters
    """
    cell = [float(x) for x in args.cell]
    no_result = '\nNo similar unit cell found.'
    if args.outfile:
        dbfilename = args.outfile
    else:
        dbfilename = 'structuredb.sqlite'
    db, structures = get_database(dbfilename)
    volume = vol_unitcell(*cell)
    atol, ltol, vol_threshold = regular_results_parameters(volume)
    # the fist number in the result is the structureid:
    cells = structures.find_by_volume(volume, vol_threshold)
    idlist = []
    if not cells:
        print(no_result)
        sys.exit()
    lattice1 = Lattice.from_parameters(*cell)
    for num, curr_cell in enumerate(cells):
        try:
            lattice2 = Lattice.from_parameters(*curr_cell[1:7])
        except ValueError:
            continue
        mapping = lattice1.find_mapping(lattice2, ltol, atol, skip_rotation_matrix=True)
        if mapping:
            idlist.append(curr_cell[0])
    if not idlist:
        print(no_result)
        sys.exit()
    else:
        print('\n{} Structures found:'.format(len(idlist)))
        searchresult = structures.get_all_structure_names(idlist)
    print(
        'ID  |      path                                                                     |   filename            |   data   ')
    print('-' * 130)
    for res in searchresult:
        Id = res[0]
        dataname, filename, path = [x.decode('utf-8') for x in res if isinstance(x, bytes)]
        print(f'{Id:3} | {path:77s} | {filename:<21s} | {dataname:s}')


def run_index(args=None):
    worker = None
    if not args:
        print('')
    else:
        if not any([args.fillres, args.fillcif]):
            print("Error: You need to give either option -c, -r or both.")
            sys.exit()
        if args.outfile:
            dbfilename = args.outfile
        else:
            dbfilename = 'structuredb.sqlite'
        if args.delete:
            try:
                dbf = Path(dbfilename)
                dbf.unlink()
            except FileNotFoundError:
                pass
            except PermissionError:
                print('Could not acess database file "{}". Is it used elsewhere?'.format(dbfilename))
                print('Giving up...')
                sys.exit()
        db, structures = get_database(dbfilename)
        time1 = time.perf_counter()
        for p in args.dir:
            # the command line version
            lastid = db.get_lastrowid()
            if not lastid:
                lastid = 1
            else:
                lastid += 1
            try:
                worker = Worker(searchpath=p, excludes=args.ex,
                                structures=structures, lastid=lastid,
                                add_res_files=args.fillres, add_cif_files=args.fillcif,
                                standalone=True)
            except OSError as e:
                print("Unable to collect files:")
                print(e)
            except KeyboardInterrupt:
                sys.exit()
            print("---------------------")
        try:
            if db and structures:
                db.init_textsearch()
                db.init_author_search()
                db.populate_fulltext_search_table()
                db.populate_author_fulltext_search()
                db.make_indexes()
        except TypeError:
            print('No valid files found. They might be in excluded subdirectories.')
        time2 = time.perf_counter()
        diff = time2 - time1
        m, s = divmod(diff, 60)
        h, m = divmod(m, 60)
        print(f"\nTotal {worker.files_indexed} cif/res files in '{str(Path(dbfilename).resolve())}'. "
              f"\nDuration: {int(h):>2d} h, {int(m):>2d} m, {s:>3.2f} s")
        import os
        if "PYTEST_CURRENT_TEST" not in os.environ:
            check_update()


def merge_database(args: Namespace):
    merge_file_name = args.merge
    dbfile = args.outfile
    if not merge_file_name or not Path(merge_file_name).is_file():
        return
    if Path(merge_file_name).samefile(dbfile):
        print('\nCan not merge same file together!\n')
        return
    db, structures = get_database(dbfile)
    db.merge_databases(merge_file_name)


def get_database(dbfilename):
    db = DatabaseRequest(dbfilename)
    try:
        db.initialize_db()
    except DatabaseError as e:
        print(e)
        print(f'The Database {dbfilename} is corrupt. Unable to open it!')
        sys.exit()
    structures = StructureTable(dbfilename)
    return db, structures


def main():
    args = parser.parse_args()
    if args.cell:
        find_cell(args)
    elif args.merge:
        merge_database(args)
    else:
        if not args.dir:
            parser.print_help()
            print("\n --> No valid search directory given.\n")
            sys.exit()
        run_index(args)


if __name__ == '__main__':
    main()
    # find_cell('10.5086  20.9035  20.5072   90.000   94.130   90.000'.split())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/Users/daniel/Documents/GitHub/StructureFinder/structurefinder/gui/strf_dbpasswd.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_PasswdDialog(object):
    def setupUi(self, PasswdDialog):
        PasswdDialog.setObjectName("PasswdDialog")
        PasswdDialog.setWindowModality(QtCore.Qt.ApplicationModal)
        PasswdDialog.resize(400, 213)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(PasswdDialog.sizePolicy().hasHeightForWidth())
        PasswdDialog.setSizePolicy(sizePolicy)
        PasswdDialog.setSizeGripEnabled(False)
        PasswdDialog.setModal(True)
        self.gridLayout = QtWidgets.QGridLayout(PasswdDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.WarnLabel = QtWidgets.QLabel(PasswdDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.WarnLabel.setFont(font)
        self.WarnLabel.setObjectName("WarnLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.SpanningRole, self.WarnLabel)
        self.Asklabel = QtWidgets.QLabel(PasswdDialog)
        self.Asklabel.setObjectName("Asklabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.SpanningRole, self.Asklabel)
        self.NameLabel = QtWidgets.QLabel(PasswdDialog)
        self.NameLabel.setObjectName("NameLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.NameLabel)
        self.userNameLineEdit = QtWidgets.QLineEdit(PasswdDialog)
        self.userNameLineEdit.setObjectName("userNameLineEdit")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.userNameLineEdit)
        self.PasswordLabel = QtWidgets.QLabel(PasswdDialog)
        self.PasswordLabel.setObjectName("PasswordLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.PasswordLabel)
        self.PasswordLineEdit = QtWidgets.QLineEdit(PasswdDialog)
        self.PasswordLineEdit.setObjectName("PasswordLineEdit")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.PasswordLineEdit)
        self.IPLabel = QtWidgets.QLabel(PasswdDialog)
        self.IPLabel.setObjectName("IPLabel")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.IPLabel)
        self.IPlineEdit = QtWidgets.QLineEdit(PasswdDialog)
        self.IPlineEdit.setObjectName("IPlineEdit")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.IPlineEdit)
        self.verticalLayout.addLayout(self.formLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(PasswdDialog)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        self.verticalLayout.addItem(spacerItem)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(PasswdDialog)
        self.buttonBox.accepted.connect(PasswdDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(PasswdDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(PasswdDialog)

    def retranslateUi(self, PasswdDialog):
        _translate = QtCore.QCoreApplication.translate
        PasswdDialog.setWindowTitle(_translate("PasswdDialog", "StructureFinder"))
        self.WarnLabel.setText(_translate("PasswdDialog", "Unable to connect to the APEX database. "))
        self.Asklabel.setText(_translate("PasswdDialog", "Please give the correct IP Adress, Username and Password."))
        self.NameLabel.setText(_translate("PasswdDialog", "Username"))
        self.userNameLineEdit.setText(_translate("PasswdDialog", "BrukerPostgreSQL"))
        self.PasswordLabel.setText(_translate("PasswdDialog", "Password"))
        self.PasswordLineEdit.setText(_translate("PasswdDialog", "Bruker-PostgreSQL"))
        self.IPLabel.setText(_translate("PasswdDialog", "IP Adress"))
        self.IPlineEdit.setText(_translate("PasswdDialog", "localhost"))



# -*- coding: utf-8 -*-
"""
Created on 09.02.2015

 ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE" (Revision 42):
* <dkratzert@gmx.de> wrote this file. As long as you retain this
* notice you can do whatever you want with this stuff. If we meet some day, and
* you think this stuff is worth it, you can buy me a beer in return.
* ----------------------------------------------------------------------------

@author: daniel
"""
from collections import namedtuple
from contextlib import suppress
from typing import List, Dict, Union, Any

import gemmi.cif
from gemmi import cif

DEBUG = False


class CifFile(object):

    def __init__(self, options=None):
        """
        A cif file parsing object optimized for speed and simplicity.
        It can not handle multi cif files.
        """
        self.doc: Union[None, gemmi.cif.Document] = None
        self.block: Union[None, gemmi.cif.Block] = None
        self.cell: Union[None, gemmi.UnitCell] = None
        if options is None:
            options = {'modification_time': "", 'file_size': ""}
        # This is a set of keys that are already there:
        self.cif_data: Dict[str, Union[str, Any]] = {
            "data"                                : '',
            "_cell_formula_units_Z"               : '',
            "_space_group_name_H-M_alt"           : '',
            "_space_group_name_Hall"              : '',
            "_space_group_centring_type"          : '',
            "_space_group_IT_number"              : '',
            "_space_group_crystal_system"         : '',
            "_space_group_symop_operation_xyz"    : '',
            "_audit_creation_method"              : '',
            "_chemical_formula_sum"               : '',
            "_chemical_formula_weight"            : '',
            "_exptl_crystal_description"          : '',
            "_exptl_crystal_colour"               : '',
            "_exptl_crystal_size_max"             : '',
            "_exptl_crystal_size_mid"             : '',
            "_exptl_crystal_size_min"             : '',
            "_exptl_absorpt_coefficient_mu"       : '',
            "_exptl_absorpt_correction_type"      : '',
            "_exptl_special_details"              : '',
            "_diffrn_ambient_temperature"         : '',
            "_diffrn_radiation_wavelength"        : '',
            "_diffrn_radiation_type"              : '',
            "_diffrn_source"                      : '',
            "_diffrn_measurement_device_type"     : '',
            "_diffrn_reflns_number"               : '',
            "_diffrn_reflns_av_R_equivalents"     : '',
            "_diffrn_reflns_av_unetI/netI"        : '',
            "_diffrn_reflns_theta_min"            : '',
            "_diffrn_reflns_theta_max"            : '',
            "_diffrn_reflns_theta_full"           : '',
            "_diffrn_measured_fraction_theta_max" : '',
            "_diffrn_measured_fraction_theta_full": '',
            "_reflns_number_total"                : '',
            "_reflns_number_gt"                   : '',
            "_reflns_threshold_expression"        : '',
            "_reflns_Friedel_coverage"            : '',
            "_computing_structure_solution"       : '',
            "_computing_structure_refinement"     : '',
            "_refine_special_details"             : '',
            "_refine_ls_abs_structure_Flack"      : '',
            "_refine_ls_structure_factor_coef"    : '',
            "'_refine_ls_hydrogen_treatment'"     : '',
            "_refine_ls_weighting_details"        : '',
            "_refine_ls_number_reflns"            : '',
            "_refine_ls_number_parameters"        : '',
            "_refine_ls_number_restraints"        : '',
            "_refine_ls_R_factor_all"             : '',
            "_refine_ls_R_factor_gt"              : '',
            "_refine_ls_wR_factor_ref"            : '',
            "_refine_ls_wR_factor_gt"             : '',
            "_refine_ls_goodness_of_fit_ref"      : '',
            "_refine_ls_restrained_S_all"         : '',
            "_refine_ls_shift/su_max"             : '',
            "_refine_ls_shift/su_mean"            : '',
            "_refine_diff_density_max"            : '',
            "_refine_diff_density_min"            : '',
            "_diffrn_reflns_av_unetI_netI"        : '',
            "_database_code_depnum_ccdc_archive"  : '',
            "_shelx_res_file"                     : '',
            "_audit_author_name"                  : '',
            "_audit_contact_author_name"          : '',
            "_citation_author_name"               : '',
            "_citation_editor_name"               : '',
            "_publ_contact_author_name"           : '',
            "_publ_contact_author"                : '',
            "_publ_author_name"                   : '',
            "modification_time"                   : options['modification_time'],
            "file_size"                           : options['file_size'],
            "calculated_formula_sum"              : '',
        }

    def parsefile(self, doc: cif.Document) -> bool:
        """
        This method parses the cif file. Currently, only single items and atoms are supported.
        :param txt: cif file as list without line endings
        :return: cif file content
        :rtype: dict
        """
        self.global_data = None
        self.doc: gemmi.cif.Document = doc
        if self.doc.find_block('global'):
            self.store_global_values()
            with suppress(IndexError):
                self.block: gemmi.cif.Block = self.doc.find_block(self.doc[1].name)
        else:
            try:
                self.block = self.doc[0]
            except IndexError:
                return False
        if not self.block:
            if DEBUG:
                print('No block found!')
            return False
        self.cell = self._cell
        if self.cell.volume < 2.0:
            return False
        if not self.cell:
            if DEBUG:
                print('No cell in cif!')
            return False
        self.fill_data_dict()
        self.update_from_global_values()
        self.handle_deprecates()
        return True

    def store_global_values(self):
        self.block = self.doc.find_block('global')
        self.fill_data_dict()
        self.global_data = self.cif_data.copy()

    def update_from_global_values(self):
        if self.global_data:
            for key, value in self.global_data.items():
                if value:
                    self.cif_data[key] = value

    def get_itnum(self):
        it_num = self.as_int('_space_group_IT_number')
        if not it_num:
            if self.space_group:
                return self.space_group.number
            else:
                return None
        return it_num

    def get_space_group(self):
        space_group = self.as_string('_space_group_name_H-M_alt')
        if space_group and space_group != '?' and space_group != '.':
            return space_group
        else:
            if self.space_group:
                return self.space_group.short_name()
            else:
                return None

    def fill_data_dict(self):
        self.cif_data['_cell_formula_units_Z'] = self['_cell_formula_units_Z']
        self.cif_data['_space_group_name_H-M_alt'] = self.get_space_group()
        self.cif_data['_space_group_name_Hall'] = self.as_string('_space_group_name_Hall')
        self.cif_data['_space_group_centring_type'] = self.as_string('_space_group_centring_type')
        self.cif_data['_space_group_IT_number'] = self.get_itnum()
        self.cif_data['_space_group_crystal_system'] = self.as_string('_space_group_crystal_system')
        self.cif_data['_space_group_symop_operation_xyz'] = '\n'.join(self.symm)
        self.cif_data['_audit_creation_method'] = self.as_string('_audit_creation_method')
        self.cif_data['_chemical_formula_sum'] = self.as_string('_chemical_formula_sum')
        self.cif_data['_chemical_formula_weight'] = self.as_number('_chemical_formula_weight')
        self.cif_data['_exptl_crystal_description'] = self.as_string('_exptl_crystal_description')
        self.cif_data['_exptl_crystal_colour'] = self.as_string('_exptl_crystal_colour')
        self.cif_data['_exptl_crystal_size_max'] = self.as_string('_exptl_crystal_size_max')
        self.cif_data['_exptl_crystal_size_mid'] = self.as_string('_exptl_crystal_size_mid')
        self.cif_data['_exptl_crystal_size_min'] = self.as_string('_exptl_crystal_size_min')
        self.cif_data['_exptl_absorpt_coefficient_mu'] = self.as_number('_exptl_absorpt_coefficient_mu')
        self.cif_data['_exptl_absorpt_correction_type'] = self.as_string('_exptl_absorpt_correction_type')
        self.cif_data['_exptl_special_details'] = self.as_string('_exptl_special_details')
        self.cif_data['_diffrn_ambient_temperature'] = self.as_number('_diffrn_ambient_temperature')
        self.cif_data['_diffrn_radiation_wavelength'] = self.as_number('_diffrn_radiation_wavelength')
        self.cif_data['_diffrn_radiation_type'] = self.as_string('_diffrn_radiation_type')
        self.cif_data['_diffrn_source'] = self.as_string('_diffrn_source')
        self.cif_data['_diffrn_measurement_device_type'] = self.as_string('_diffrn_measurement_device_type')
        self.cif_data['_diffrn_reflns_number'] = self.as_string('_diffrn_reflns_number')
        self.cif_data['_diffrn_reflns_av_R_equivalents'] = self.as_string('_diffrn_reflns_av_R_equivalents')
        self.cif_data['_diffrn_reflns_av_unetI/netI'] = self.as_string('_diffrn_reflns_av_unetI/netI')
        self.cif_data['_diffrn_reflns_theta_min'] = self.as_string('_diffrn_reflns_theta_min')
        self.cif_data['_diffrn_reflns_theta_max'] = self.as_string('_diffrn_reflns_theta_max')
        self.cif_data['_diffrn_reflns_theta_full'] = self.as_string('_diffrn_reflns_theta_full')
        self.cif_data['_diffrn_measured_fraction_theta_max'] = self.as_number('_diffrn_measured_fraction_theta_max')
        self.cif_data['_diffrn_measured_fraction_theta_full'] = self.as_number('_diffrn_measured_fraction_theta_full')
        self.cif_data['_reflns_number_total'] = self.as_string('_reflns_number_total')
        self.cif_data['_reflns_number_gt'] = self.as_string('_reflns_number_gt')
        self.cif_data['_reflns_threshold_expression'] = self.as_string('_reflns_threshold_expression')
        self.cif_data['_reflns_Friedel_coverage'] = self.as_number('_reflns_Friedel_coverage')
        self.cif_data['_computing_structure_solution'] = self.as_string('_computing_structure_solution')
        self.cif_data['_computing_structure_refinement'] = self.as_string('_computing_structure_refinement')
        self.cif_data['_refine_special_details'] = self.as_string('_refine_special_details')
        self.cif_data['_refine_ls_abs_structure_Flack'] = self.as_number('_refine_ls_abs_structure_Flack')
        self.cif_data['_refine_ls_structure_factor_coef'] = self.as_string('_refine_ls_structure_factor_coef')
        self.cif_data['_refine_ls_hydrogen_treatment'] = self.as_string('_refine_ls_hydrogen_treatment')
        self.cif_data['_refine_ls_weighting_details'] = self.as_string('_refine_ls_weighting_details')
        self.cif_data['_refine_ls_number_reflns'] = self.as_string('_refine_ls_number_reflns')
        self.cif_data['_refine_ls_number_parameters'] = self.as_int('_refine_ls_number_parameters')
        self.cif_data['_refine_ls_number_restraints'] = self.as_int('_refine_ls_number_restraints')
        self.cif_data['_refine_ls_R_factor_all'] = self.as_number('_refine_ls_R_factor_all')
        self.cif_data['_refine_ls_R_factor_gt'] = self.as_number('_refine_ls_R_factor_gt')
        self.cif_data['_refine_ls_wR_factor_ref'] = self.as_number('_refine_ls_wR_factor_ref')
        self.cif_data['_refine_ls_wR_factor_gt'] = self.as_number('_refine_ls_wR_factor_gt')
        self.cif_data['_refine_ls_goodness_of_fit_ref'] = self.as_number('_refine_ls_goodness_of_fit_ref')
        self.cif_data['_refine_ls_restrained_S_all'] = self.as_number('_refine_ls_restrained_S_all')
        self.cif_data['_refine_ls_shift/su_max'] = self.as_number('_refine_ls_shift/su_max')
        self.cif_data['_refine_ls_shift/su_mean'] = self.as_number('_refine_ls_shift/su_mean')
        self.cif_data['_refine_diff_density_max'] = self.as_number('_refine_diff_density_max')
        self.cif_data['_refine_diff_density_min'] = self.as_number('_refine_diff_density_min')
        self.cif_data['_diffrn_reflns_av_unetI_netI'] = self.as_string('_diffrn_reflns_av_unetI_netI')
        self.cif_data['_database_code_depnum_ccdc_archive'] = self.as_string('_database_code_depnum_ccdc_archive')
        self.cif_data['_shelx_res_file'] = self.as_string('_shelx_res_file')
        if not self.cif_data['_shelx_res_file']:
            self.cif_data['_shelx_res_file'] = self.as_string('_iucr_refine_instructions_details')
        self.cif_data['_audit_author_name'] = self.as_list('_audit_author_name')
        self.cif_data['_audit_contact_author_name'] = self.as_list('_audit_contact_author_name')
        self.cif_data['_citation_author_name'] = self.as_list('_citation_author_name')
        self.cif_data['_citation_editor_name'] = self.as_list('_citation_editor_name')
        self.cif_data['_publ_contact_author_name'] = self.as_list('_publ_contact_author_name')
        self.cif_data['_publ_contact_author'] = self.as_list('_publ_contact_author')
        self.cif_data['_publ_author_name'] = self.as_list('_publ_author_name')

    def as_list(self, cif_key: str):
        try:
            return '\n'.join([cif.as_string(x) for x in self.block.find_values(cif_key)])
        except UnicodeDecodeError:
            return ''

    def as_string(self, cif_key: str) -> str:
        return cif.as_string(self[cif_key]) if self[cif_key] else ''

    def as_int(self, cif_key: str) -> Union[int, None]:
        if self[cif_key]:
            try:
                return cif.as_int(self[cif_key])
            except ValueError as e:
                if DEBUG:
                    print('Failed to convert to int:')
                    print(e, self[cif_key])
                    # raise
                return None
        else:
            return None

    def as_number(self, cif_key: str) -> Union[float, None]:
        return cif.as_number(self[cif_key]) if self[cif_key] else ''

    def handle_deprecates(self):
        """
        Makes the old and new cif values equal.
        """
        if self["_symmetry_space_group_name_H-M"]:
            self.cif_data["_space_group_name_H-M_alt"] = self.as_string("_symmetry_space_group_name_H-M")
        if self["_diffrn_measurement_device"]:
            self.cif_data["_diffrn_measurement_device_type"] = self.as_string("_diffrn_measurement_device")
        if self["_refine_ls_shift/esd_max"]:
            self.cif_data["_refine_ls_shift/su_max"] = self.as_number("_refine_ls_shift/esd_max")
        if self['_symmetry_space_group_name_Hall']:
            self.cif_data['_space_group_name_Hall'] = self.as_string('_symmetry_space_group_name_Hall')
        if self['_symmetry_Int_Tables_number']:
            self.cif_data['_space_group_IT_number'] = self.as_int('_symmetry_Int_Tables_number')
        if self['_diffrn_reflns_av_sigmaI/netI']:
            self.cif_data['_diffrn_reflns_av_unetI/netI'] = self.as_number('_diffrn_reflns_av_sigmaI/netI')

        if self.as_string("_space_group_name_H-M_alt") and not self.as_string('_space_group_centring_type') or \
            self['_space_group_centring_type'] == '?':
            try:
                self.cif_data["_space_group_centring_type"] = self.as_string("_space_group_name_H-M_alt").split()[0][0]
            except IndexError as e:
                print(self.as_string("_space_group_name_H-M_alt"))
                raise
        elif self.as_string('_space_group_name_Hall') and not self.as_string('_space_group_centring_type'):
            try:
                self.cif_data["_space_group_centring_type"] = \
                    self.as_string('_space_group_name_Hall').split()[0].lstrip('-')[0]
            except IndexError:
                pass
        else:
            self.cif_data["_space_group_centring_type"] = self.space_group.centring_type() if self.space_group else '?'
        if self['_symmetry_cell_setting']:
            self.cif_data['_space_group_crystal_system'] = self.as_string('_symmetry_cell_setting')

    def __getitem__(self, item) -> str:
        """
        Returns an attribute of the cif data dictionary.
        """
        try:
            return self.block.find_value(item)
        except (KeyError, UnicodeDecodeError):
            return ''

    @property
    def _cell(self) -> gemmi.UnitCell:
        a = self['_cell_length_a']
        b = self['_cell_length_b']
        c = self['_cell_length_c']
        alpha = self['_cell_angle_alpha']
        beta = self['_cell_angle_beta']
        gamma = self['_cell_angle_gamma']
        if not all((a, b, c, alpha, beta, gamma)):
            return gemmi.UnitCell()
        return gemmi.UnitCell(cif.as_number(a), cif.as_number(b), cif.as_number(c),
                              cif.as_number(alpha), cif.as_number(beta), cif.as_number(gamma))

    @property
    def volume(self) -> float:
        return self.cell.volume

    @property
    def space_group(self) -> Union[gemmi.SpaceGroup, None]:
        try:
            gops = gemmi.GroupOps([gemmi.Op(o) for o in self.symm])
            spgr = gemmi.find_spacegroup_by_ops(gops)
        except RuntimeError as e:
            if DEBUG:
                print('Space group not found:', e)
            return None
        return spgr

    @property
    def atoms(self):
        """
        Atoms from the CIF where values are returned as string like in the CIF with esds.
        """
        labels: List[str] = self.block.find_loop('_atom_site_label')
        types: List[str] = self.block.find_loop('_atom_site_type_symbol')
        x = self.block.find_loop('_atom_site_fract_x')
        y = self.block.find_loop('_atom_site_fract_y')
        z = self.block.find_loop('_atom_site_fract_z')
        part = self.block.find_loop('_atom_site_disorder_group')
        occ = self.block.find_loop('_atom_site_occupancy')
        u_eq = self.block.find_loop('_atom_site_U_iso_or_equiv')
        atom = namedtuple('Atom', ('label', 'type', 'x', 'y', 'z', 'part', 'occ', 'u_eq'))
        for label, type, x, y, z, part, occ, u_eq in zip(labels, types, x, y, z,
                                                         part if part else ('0',) * len(labels),
                                                         occ if occ else ('1.000000',) * len(labels),
                                                         u_eq):
            yield atom(label=label, type=type, x=x, y=y, z=z, part=part, occ=occ, u_eq=u_eq)

    @property
    def atoms_orth(self):
        atom = namedtuple('Atom', ('label', 'type', 'x', 'y', 'z', 'part', 'occ', 'u_eq'))
        for at in self.atoms:
            x, y, z = self.cell.orthogonalize(
                gemmi.Fractional(cif.as_number(at.x), cif.as_number(at.y), cif.as_number(at.z)))
            yield atom(label=at.label, type=at.type, x=x, y=y, z=z,
                       part=at.part, occ=at.occ, u_eq=at.u_eq)

    @property
    def symm(self) -> List[str]:
        """
        Yields symmetry operations.
        ['x, y, z', '-x+1/2, y+1/2, -z+1/2', '-x, -y, -z', 'x-1/2, -y-1/2, z-1/2']
        """
        symm1 = self.block.find_values('_space_group_symop_operation_xyz')
        symm2 = self.block.find_values('_symmetry_equiv_pos_as_xyz')
        return [cif.as_string(x) for x in symm1] if symm1 else [cif.as_string(x) for x in symm2]


if __name__ == '__main__':
    fullpath = r'./tests/test-data/COD/4060314.cif'
    doc = gemmi.cif.Document()
    doc.source = fullpath
    doc.parse_file(fullpath)
    c = CifFile()
    c.parsefile(doc)
    print(list(c.symm))

    # c = CifFile()
    # cifok = c.parsefile(r'/Users/daniel/Downloads/j9uwm9c3tmp.cif')
    # pprint(cif.cif_data)
    # pprint(cif._space_group_centring_type)
    # sys.exit()
    # TODO: "_space_group_symop_operation_xyz" or '_symmetry_equiv_pos_as_xyz':
    # print(len(list([x for x in c.atoms])))
    #    pass
    #    print(x)
    # print(cifok)
    # import doctest

    # failed, attempted = doctest.testmod()  # verbose=True)
    # if failed == 0:
    #    print('passed all {} tests!'.format(attempted))



import os
import subprocess
import sys
import xml.etree.ElementTree as ET
from pathlib import Path
from pprint import pprint
from shutil import which
from tempfile import mkstemp
from typing import Union

from pkg_resources import parse_version

DEBUG = False

try:
    from winreg import OpenKey, HKEY_CURRENT_USER, EnumKey, QueryInfoKey, EnumValue
except ImportError:
    pass

querytext = """<?xml version="1.0" encoding="UTF-8"?>
<query name="reduced_cell_search" version="1.0" originator="generic">
  <lattice_centring></lattice_centring>
  <a></a>
  <b></b>
  <c></c>
  <alpha></alpha>
  <beta></beta>
  <gamma></gamma>
  <settings>
    <dimension_tolerance>1.5</dimension_tolerance>
    <angle_tolerance>2.0</angle_tolerance>
    <maximum_hits>2000</maximum_hits>
  </settings>
</query> """


def get_cccsd_path() -> Union[Path, None]:
    r"""
    HKEY_CURRENT_USER\SOFTWARE\CCDC\CellCheckCSD
    """
    software = r'SOFTWARE\CCDC\CellCheckCSD\\'
    try:
        csd = OpenKey(HKEY_CURRENT_USER, software)
    except (FileNotFoundError, NameError):
        return None
    num = QueryInfoKey(csd)[0]  # returns the number of subkeys
    csd_path = None
    latest = parse_version('0.0')
    for n in range(num):
        vernum_dir = EnumKey(csd, n)  # subkey of version
        path = OpenKey(HKEY_CURRENT_USER, software + vernum_dir)
        ver = parse_version(vernum_dir)
        if QueryInfoKey(path)[1] > 0:  # subkey with content
            if latest < ver:
                csd_path = Path(EnumValue(path, 0)[1], 'ccdc_searcher.bat')
    if csd_path and csd_path.is_file():
        return csd_path
    else:
        try:
            if which('ccdc_searcher.bat'):
                return Path(which('ccdc_searcher.bat'))
            else:
                return Path(which('ccdc_searcher'))
        except TypeError:
            return None


def read_queryfile():
    """
    Reads a CellCheckCSD query file.
    """
    tree = ET.fromstring(querytext)
    return tree


def set_unitcell(cell: list, centering: str) -> ET:
    """
    Returns an xml query string for the CellCheckcsd -query option.
    """
    t = read_queryfile()
    cent = t.find('lattice_centring')
    a = t.find('a')
    b = t.find('b')
    c = t.find('c')
    alpha = t.find('alpha')
    beta = t.find('beta')
    gamma = t.find('gamma')
    a.text = str(cell[0])
    b.text = str(cell[1])
    c.text = str(cell[2])
    alpha.text = str(cell[3])
    beta.text = str(cell[4])
    gamma.text = str(cell[5])
    cent.text = centering
    return ET.tostring(t)


def parse_results(xmlinput: str) -> list:
    """
    Parses the search results into a dictionary.
    """
    root = ET.fromstring(xmlinput)
    results = []
    # each hit has a match:
    for num, r in enumerate(root.findall('match')):
        # The identifier is the csd entry name:
        ident = r.get('identifier')
        values = {'recid': ident}
        values['order'] = num
        # unit cells are listed in parameters:
        for p in r.findall('parameters'):
            for item in p.findall('parameter'):
                item_type = item.get('type')
                value = item.text
                values[item_type] = value
        results.append(values)
    return results


def search_csd(cell: list, centering: str, searcher_executable: str = '') -> str:
    querystring = set_unitcell(cell, centering).decode('utf-8')
    # prepare the temporary queryfile:
    querydescriptor, queryfile = mkstemp(suffix='xml')
    # prepare the temporary results file:
    resdescriptor, resultfile = mkstemp(suffix='xml')
    # write query to file:
    Path(queryfile).write_text(querystring)
    if sys.platform == 'win32':
        shell = True
    else:
        shell = False
    # Either use the path from the text field or from get_cccsd_path()
    if searcher_executable and Path(searcher_executable).exists() and Path(searcher_executable).is_file():
        path_to_exe = Path(searcher_executable)
    else:
        path_to_exe = get_cccsd_path()
    result_file = ''
    try:
        if DEBUG:
            print(f'Running CCDC query: {path_to_exe}')
        # run the search:
        # shell=True disables the output window of the process
        subprocess.run([str(path_to_exe.absolute()), '-query', queryfile, '-results', resultfile], shell=shell)
        # read the result:
        result_file = Path(resultfile).read_text(encoding='utf-8', errors='ignore')
    except Exception as e:
        print('Could not search cell:')
        print(e)
    os.close(resdescriptor)
    os.close(querydescriptor)
    try:
        os.remove(queryfile)
        os.remove(resultfile)
    except Exception:
        pass
    return result_file


if __name__ == '__main__':
    # print(get_cccsd_path())
    cell = [10.6369, 10.6369, 24.5938, 90.0, 90.0, 120.0]
    xml = search_csd(cell, centering='R')
    res = parse_results(xml)
    pprint(res)





from math import acos, sqrt, degrees

from structurefinder.shelxfile import elements
from structurefinder.shelxfile.cards import AFIX, PART, RESI
from structurefinder.shelxfile.dsrmath import atomic_distance, frac_to_cart, Array
from structurefinder.shelxfile.misc import DEBUG, split_fvar_and_parameter, ParseUnknownParam, ParseSyntaxError

"""
TODO:

"""


class Atoms():
    """
    All atoms from a SHELXL file with their properties.
    """

    def __init__(self, shx):
        self.shx = shx
        self.all_atoms = []

    def append(self, atom: 'Atom') -> None:
        """
        Adds a new atom to the list of atoms. Using append is essential.
        """
        self.all_atoms.append(atom)

    @property
    def nameslist(self):
        return [at.fullname.upper() for at in self.all_atoms]

    def __repr__(self):
        if self.all_atoms:
            return '\n'.join([str(x) for x in self.all_atoms])
        else:
            return 'No Atoms in file.'

    def __iter__(self):
        return iter(x for x in self.all_atoms)

    def __getitem__(self, item: int) -> 'Atom':
        return self.get_atom_by_id(item)

    def __len__(self) -> int:
        return len(self.all_atoms)

    def __delitem__(self, key):
        """
        Delete an atom by its atomid:
        del atoms[4]
        """
        for n, at in enumerate(self.all_atoms):
            if key == at.atomid:
                if DEBUG:
                    print("deleting atom", at.fullname)
                del self.all_atoms[n]
                del self.shx._reslist[self.shx._reslist.index(at)]
        # if DEBUG:
        #    print('Could not delete atom {}'.format(self.get_atom_by_id(key.atomid).fullname))

    @property
    def atomsdict(self):
        return dict((atom.fullname, atom) for atom in self.all_atoms)

    @property
    def number(self) -> int:
        """
        The number of atoms in the current SHELX file.
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> shx.atoms.number
        148
        """
        return len(self.all_atoms)

    def get_atom_by_id(self, aid: int) -> 'Atom':
        """
        Returns the atom objext with atomId id.
        """
        for a in self.all_atoms:
            if aid == a.atomid:
                return a

    def has_atom(self, atom_name: str) -> bool:
        """
        Returns true if shelx file has atom.

        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> shx.atoms.has_atom('Al1')
        True
        >>> shx.atoms.has_atom('Al1_0')
        True
        >>> shx.atoms.has_atom('Al2_0')
        False
        """
        if '_' not in atom_name:
            atom_name += '_0'
        if atom_name.upper() in self.nameslist:
            return True
        else:
            return False

    def get_atom_by_name(self, atom_name: str) -> 'Atom' or None:
        """
        Returns an Atom object using an atom name with residue number like C1, C1_0, F2_4, etc.
        C1 means atom C1 in residue 0.
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> shx.atoms.get_atom_by_name('Al1')
        Atom ID: 73
        """
        if '_' not in atom_name:
            if atom_name == ">" or atom_name == "<":
                return None
            atom_name += '_0'
        try:
            at = self.atomsdict[atom_name.upper()]
        except KeyError:
            print("Atom {} not found in atom list.".format(atom_name))
            return None
        return at

    def get_multi_atnames(self, atom_name, residue_class):
        atoms = []
        if residue_class:
            for num in self.shx.residues.residue_classes[residue_class]:
                if '_' not in atom_name:
                    atom_name += '_0'
                else:
                    atom_name += '_{}'.format(num)
                try:
                    atoms.append(self.atomsdict[atom_name.upper()])
                except KeyError:
                    pass
        else:
            try:
                atoms.append(self.atomsdict[atom_name.upper()])
            except KeyError:
                return None
        return atoms

    def get_all_atomcoordinates(self) -> dict:
        """
        Returns a dictionary {'C1': ['1.123', '0.7456', '3.245'], 'C2_2': ...}
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> shx.atoms.get_all_atomcoordinates() # doctest: +ELLIPSIS
        {'O1_4': [0.074835, 0.238436, 0.402457], 'C1_4': [0.028576, 0.234542, 0.337234], ...}
        """
        atdict = {}
        for at in self.all_atoms:
            # if at.qpeak:
            #    atdict[at.name] = at.frac_coords
            # else:
            atdict[at.name.upper() + '_' + str(at.resinum)] = at.frac_coords
        return atdict

    def get_frag_fend_atoms(self) -> list:
        """
        Returns a list of atoms with cartesian coordinates. Atom names and sfac are ignored. They come from AFIX 17x.
        [[0.5316439256202359, 7.037351406500001, 10.112963255220803],
        [-1.7511017452002604, 5.461541059000001, 10.01187984858907]]
        """
        atoms = []
        for at in self.all_atoms:
            if at.frag_atom:
                atoms.append([at.xc, at.yc, at.zc])
        return atoms

    @property
    def residues(self) -> list:
        """
        Returns a list of the residue numbers in the shelx file.
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> shx.atoms.residues
        [0, 1, 2, 3, 4]
        """
        return list(set([x.resinum for x in self.all_atoms]))

    @property
    def q_peaks(self) -> list:
        r"""
        Returns a list of q-peaks in the file.
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> shx.atoms.q_peaks[:5] # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        [Atom ID: 328, Atom ID: 329, Atom ID: 330, Atom ID: 331, Atom ID: 332]
        """
        return [x for x in self.all_atoms if x.qpeak]

    def distance(self, atom1: str, atom2: str) -> float:
        """
        Calculates the (shortest) distance of two atoms given as text names e.g. C1_3.
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> round(shx.atoms.distance('F1_2', 'F2_2'), 6)
        2.154399
        >>> round(shx.atoms.distance('C2_2', 'F1_2'), 6)
        1.332854
        """
        a1 = self.get_atom_by_name(atom1)
        a2 = self.get_atom_by_name(atom2)
        try:
            return atomic_distance([a1.xc, a1.yc, a1.zc], [a2.xc, a2.yc, a2.zc])
        except AttributeError:
            return 0.0

    def angle(self, at1: 'Atom', at2: 'Atom', at3: 'Atom') -> float:
        """
        Calculates the angle between three atoms.

        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> at1 = shx.atoms.get_atom_by_name('O1_4')
        >>> at2 = shx.atoms.get_atom_by_name('C1_4')
        >>> at3 = shx.atoms.get_atom_by_name('C2_4')
        >>> round(shx.atoms.angle(at1, at2, at3), 6)
        109.688123
        """
        ac1 = Array(at1.cart_coords)
        ac2 = Array(at2.cart_coords)
        ac3 = Array(at3.cart_coords)
        vec1 = ac2 - ac1
        vec2 = ac2 - ac3
        return vec1.angle(vec2)

    def torsion_angle(self, at1: 'Atom', at2: 'Atom', at3: 'Atom', at4: 'Atom') -> float:
        """
        Calculates the torsion angle (dieder angle) between four atoms.

        From the book of Camelo Giacovazzo:
        For a sequence of four atoms A, B, C, D, the torsion angle w(ABCD) is
        defined as the angle between the normals to the planes ABC and BCD.
        By convention w is positive if the sense of rotation from BA to
        CD, viewed down BC, is clockwise, otherwise it is negative.

        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> at1 = shx.atoms.get_atom_by_name('O1')
        >>> at2 = shx.atoms.get_atom_by_name('C1')
        >>> at3 = shx.atoms.get_atom_by_name('C2')
        >>> at4 = shx.atoms.get_atom_by_name('F1')
        >>> round(shx.atoms.torsion_angle(at1, at2, at3, at4), 6)
        74.095731
        >>> at4 = shx.atoms.get_atom_by_name('F2')
        >>> round(shx.atoms.torsion_angle(at1, at2, at3, at4), 6)
        -44.467358
        """
        ac1 = Array(at1.cart_coords)
        ac2 = Array(at2.cart_coords)
        ac3 = Array(at3.cart_coords)
        ac4 = Array(at4.cart_coords)
        # Three vectors between four atoms:
        v1 = ac2 - ac1
        v2 = ac3 - ac2
        v3 = ac4 - ac3
        # cross product:
        a = v1.cross(v2)
        b = v2.cross(v3)
        # If direction > 0, angle is positive, else negative:
        direction = v1[0] * v2[1] * v3[2] - v1[2] * v1[1] * v3[0] + v1[2] * v2[0] * v3[1] - v1[0] \
                    * v2[2] * v3[1] + v1[1] * v2[2] * v3[0] - v1[1] * v2[0] * v3[2]
        # angle between plane normals:
        ang = acos((a[0] * b[0] + a[1] * b[1] + a[2] * b[2]) /
                   (sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]) *
                    sqrt(b[0] * b[0] + b[1] * b[1] + b[2] * b[2])))
        return degrees(ang) if direction > 0 else degrees(-ang)

    def atoms_in_class(self, name: str) -> list:
        """
        Returns a list of atoms in residue class 'name'
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> shx.atoms.atoms_in_class('CCF3')
        ['O1', 'C1', 'C2', 'F1', 'F2', 'F3', 'C3', 'F4', 'F5', 'F6', 'C4', 'F7', 'F8', 'F9']
        """
        atoms = []
        for x in self.all_atoms:
            if x.resiclass == name:
                if x.name not in atoms:
                    atoms.append(x.name)
        return atoms


class Atom():
    """
    An object holding all Properties of a shelxl atom plus some extra information like
    kartesian coordinates and element type.
    """
    #                name    sfac     x         y        z       occ      u11      u12 ...
    _anisatomstr = '{:<4.4s}{:>3}{:>12.6f}{:>12.6f}{:>12.6f}{:>12.5f}{:>11.5f}=\n  {:>11.5f}' \
                   ' {:>12.5f}{:>11.5f}{:>11.5f}{:>11.5f}'  # Line wrap is handled during file write.
    #               name    sfac     x         y         z         occ      u11
    _isoatomstr = '{:<5.5s} {:<3}{:>10.6f}  {:>10.6f}  {:>9.6f}  {:>9.5f}  {:>9.5f}'
    _qpeakstr = '{:<5.5s} {:<3}{:>8.4f}  {:>8.4f}  {:>8.4f}  {:>9.5f}  {:<9.2f} {:<9.2f}'
    _fragatomstr = '{:<5.5s} {:>10.6f}  {:>10.6f}  {:>9.6f}'

    def __init__(self, shx) -> None:
        self.shx = shx
        self.cell = shx.cell
        self.sfac_num = None
        self.resi = None
        self.part = None
        self.afix = None
        self.name = 'name'  # Name without residue number like "C1"
        # Site occupation factor including free variable like 31.0
        self.sof = 11.0
        # fractional coordinates:
        self.x = None
        self.y = None
        self.z = None
        # cartesian coordinates:
        self.xc = None
        self.yc = None
        self.zc = None
        self.qpeak = False
        self.peak_height = 0.0
        self.uvals = [0.04]  # [U] or [u11 u12 u13 u21 u22 u23]
        self.uvals_orig = [0.04]
        self.frag_atom = False
        self.restraints = []
        self.previous_non_h = None  # Find in self.shx.atoms durinf initialization
        self._line_numbers = None
        self._occupancy = 1.0
        self.molindex = 0
        # Indicates if this atom is generated by symmetry:
        self.symmgen = False

    @property
    def atomid(self) -> int:
        if self.symmgen:
            return 0
        try:
            return self.shx._reslist.index(self)
        except ValueError:
            return 0

    @property
    def fullname(self) -> str:
        return self.name + '_' + str(self.resinum)  # Name including residue nimber like "C1_2"

    @property
    def resiclass(self) -> str:
        return self.resi.residue_class

    @property
    def resinum(self) -> int:
        return self.resi.residue_number

    @property
    def chain_id(self) -> str:
        return self.resi.chainID

    @property
    def fvar(self) -> int:
        # Be aware: fvar can be negative!
        fvar, _ = split_fvar_and_parameter(self.sof)
        return fvar

    @property
    def occupancy(self) -> float:
        # Only the occupancy of the atom like 0.5 (without the free variable)
        _, occ = split_fvar_and_parameter(self.sof)
        # Fractional occupancy:
        if abs(self.fvar) == 1:
            return occ
        else:
            if occ > 0:
                try:
                    occ = self.shx.fvars[self.fvar] * occ
                except IndexError:
                    occ = 1.0  # Happens if the self.fvar is not defined
                    if DEBUG:
                        raise ParseSyntaxError
            else:
                try:
                    occ = 1 + (self.shx.fvars[self.fvar] * occ)
                except IndexError:
                    occ = 1.0
                    if DEBUG:
                        raise ParseSyntaxError
        return occ

    @occupancy.setter
    def occupancy(self, occ: float):
        self._occupancy = occ

    @property
    def ishydrogen(self) -> bool:
        """
        Returns True if the current atom is a hydrogen isotope.
        """
        if self.element in ['H', 'D', 'T']:
            return True
        else:
            return False

    def set_atom_parameters(self, name: str = 'C', sfac_num: int = 1, coords: list = None, part: PART = None,
                            afix: AFIX = None, resi: RESI = None, site_occupation: float = 11.0, uvals: list = None,
                            symmgen: bool = True):
        """
        Sets atom properties manually if not parsed from a SHELXL file.
        """
        self.name = name
        self.sfac_num = sfac_num
        self.frac_coords = coords
        self.x, self.y, self.z = coords[0], coords[1], coords[2]
        self.xc, self.yc, self.zc = frac_to_cart(self.frac_coords, self.cell)
        self.part = part
        self.afix = afix
        self.resi = resi
        self.sof = site_occupation
        self.uvals = uvals
        self.symmgen = symmgen

    def set_uvals(self, uvals: list):
        """
        Sets u values and checks if a free variable was used.
        """
        self.uvals = uvals
        if len(uvals) != 2:  # two means Uiso anf q-peak hight
            for n, u in enumerate(uvals):
                if abs(u) > 4.0:
                    fvar, uval = split_fvar_and_parameter(u)
                    #self.uvals[n] = uval
                    self.shx.fvars.set_fvar_usage(fvar)
        else:
            if abs(uvals[0]) > 4.0:
                fvar, uval = split_fvar_and_parameter(uvals[0])
                self.shx.fvars.set_fvar_usage(fvar)

    def parse_line(self, atline: list, list_of_lines: list, part: PART, afix: AFIX, resi: RESI):
        """
        Parsers the text line of an atom from SHELXL to initialize the atom parameters.
        """
        self.name = atline[0][:4]  # Atom names are limited to 4 characters
        uvals = [float(x) for x in atline[6:12]]
        self.uvals_orig = uvals[:]
        self.set_uvals(uvals)
        self._line_numbers = list_of_lines
        self.part = part
        self.afix = afix
        self.resi = resi
        # TODO: test all variants of PART and AFIX sof combinations:
        if self.part.sof != 11.0:
            if self.afix and self.afix.sof:  # handles position of afix and part:
                if self.afix.index > self.part.index:
                    self.sof = self.afix.sof
            else:
                self.sof = self.part.sof
        elif self.afix and self.afix.sof:
            if self.part.sof != 11.0:
                if self.part.index > self.afix.index:
                    self.sof = self.part.sof
            else:
                self.sof = self.afix.sof
        else:
            self.sof = float(atline[5])
        try:
            x, y, z = [float(x) for x in atline[2:5]]
        except ValueError as e:
            if DEBUG:
                print(e, 'Line:', self._line_numbers[-1])
            raise ParseUnknownParam
        if abs(x) > 4:
            fvar, x = split_fvar_and_parameter(x)
            self.shx.fvars.set_fvar_usage(fvar)
        if abs(y) > 4:
            fvar, x = split_fvar_and_parameter(y)
            self.shx.fvars.set_fvar_usage(fvar)
        if abs(z) > 4:
            fvar, x = split_fvar_and_parameter(z)
            self.shx.fvars.set_fvar_usage(fvar)
        self.x = x
        self.y = y
        self.z = z
        self.xc, self.yc, self.zc = frac_to_cart(self.frac_coords, self.cell)
        if len(self.uvals) == 2 and self.shx.hklf:  # qpeaks are always behind hklf
            self.peak_height = uvals.pop()
            self.qpeak = True
        if self.shx.end:  # After 'END' can only be Q-peaks!
            self.qpeak = True
        self.sfac_num = int(atline[1])
        #self.shx.fvars.set_fvar_usage(self.fvar)

    @property
    def element(self) -> str:
        """
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('test-data/p21c.res')
        >>> at = shx.atoms.get_atom_by_name('C1_4')
        >>> at.sfac_num
        1
        >>> at.element
        'C'
        >>> at.element = 'O'
        >>> at.element
        'O'
        >>> at.sfac_num
        3
        """
        return self.shx.sfac2elem(self.sfac_num).capitalize()

    @property
    def an(self):
        return elements.get_atomic_number(self.element)

    @element.setter
    def element(self, new_element: str) -> None:
        """
        Sets the element type of an atom.
        """
        self.sfac_num = self.shx.elem2sfac(new_element)

    @property
    def radius(self) -> float:
        """
        Returns the atomic covalence radius in angstrom.
        """
        return elements.get_radius_from_element(self.element)

    def __iter__(self):
        for x in self.__repr__().split():
            yield x

    def __repr__(self) -> str:
        return 'Atom ID: ' + str(self.atomid)

    def __str__(self) -> str:
        """
        Returns a text line of the Atom with SHELXL syntax.
        :return: SHELX-formated atom string
        """
        if self.afix and self.shx.frag:
            # An atom from a FRAG/FEND instruction
            return Atom._fragatomstr.format(self.name, self.x, self.y, self.z)
        else:
            if len(self.uvals) > 2:
                # anisotropic atom
                try:
                    return Atom._anisatomstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof,
                                                    *self.uvals)
                except IndexError:
                    return 'REM Error in U values.'
            else:
                # isotropic atom:
                if self.qpeak:
                    return Atom._qpeakstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof, 0.04,
                                                 self.peak_height)
                try:
                    return Atom._isoatomstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof,
                                                   *self.uvals)
                except IndexError:
                    return Atom._isoatomstr.format(self.name, self.sfac_num, self.x, self.y, self.z, self.sof, 0.04)

    def resolve_restraints(self):
        """
        This method should generate a list of restraints objects for each restraints involved with this atom.
        TODO: Make this work
        """
        for num, r in enumerate(self.shx.restraints):
            for at in r.atoms:
                # print(r.residue_number, self.resinum, r.residue_class, self.resiclass, self.name, at)
                if r.residue_number == self.resinum and r.residue_class == self.resiclass and self.name == at:
                    self.restraints.append(r)

    @property
    def index(self):
        # The position in the res file as index number (starting from 0).
        return self.shx.index_of(self)

    @property
    def frac_coords(self, rounded=False) -> tuple:
        if rounded:
            return (round(self.x, 14), round(self.y, 14), round(self.z, 14))
        else:
            return (self.x, self.y, self.z)

    @frac_coords.setter
    def frac_coords(self, coords: list):
        self.x, self.y, self.z = coords

    @property
    def cart_coords(self):
        return [round(self.xc, 14), round(self.yc, 14), round(self.zc, 14)]

    def delete(self):
        """
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> at = shx.atoms.get_atom_by_id(40)
        >>> shx.atoms.all_atoms[:3]
        [Atom ID: 38, Atom ID: 40, Atom ID: 42]
        >>> at.fullname
        'C1_4'
        >>> at.delete()
        >>> shx.atoms.all_atoms[:3]
        [Atom ID: 38, Atom ID: 41, Atom ID: 43]
        """
        del self.shx.atoms[self.index]

    def to_isotropic(self) -> None:
        """
        Makes the current atom isotropic.
        """
        self.uvals = [0.04]

    '''
    def __eq__(self, other) -> bool:
        """
        Returns True if two atoms are of same name, part and residue.
        """
        if isinstance(other, str):
            return self.__repr__() == other
        if self.atomid == other.atomid:
            return True
        else:
            return False
    '''

    def find_atoms_around(self, dist=1.2, only_part=0) -> list:
        """
        Finds atoms around the current atom.
        >>> from structurefinder import ShelXFile
        >>> shx = ShelXFile('./test-data/p21c.res')
        >>> at = shx.atoms.get_atom_by_name('C1_4')
        >>> at.find_atoms_around(dist=2, only_part=2)
        [Atom ID: 38, Atom ID: 42, Atom ID: 50, Atom ID: 58]
        >>> shx.atoms.get_atom_by_name('C1_4').cart_coods
        [-0.19777464582151, 4.902748697, 6.89776640065679]
        """
        found = []
        for at in self.shx.atoms:
            if atomic_distance([self.x, self.y, self.z], [at.x, at.y, at.z], self.cell) < dist:
                # Not the atom itselv:
                if not self == at:
                    # only in special part and no q-peaks:
                    if at.part.n == only_part and not at.qpeak:
                        found.append(at)
        return found



from pathlib import Path

from PyQt5.QtCore import QSettings


class StructureFinderSettings():
    def __init__(self):
        self.software_name = 'StructureFinder'
        self.organization = 'DK'
        self.settings = QSettings(self.organization, self.software_name)
        self.settings.setDefaultFormat(QSettings.IniFormat)

    def save_current_work_dir(self, dir: str) -> None:
        """
        Saves the last directory where we saved a database.
        """
        self.settings.beginGroup("WorkDir")
        self.settings.setValue('dir', dir)
        self.settings.endGroup()

    def load_last_workdir(self) -> str:
        self.settings.beginGroup('WorkDir')
        lastdir = self.settings.value("dir", type=str)
        if not Path(lastdir).exists():
            lastdir = './'
        self.settings.endGroup()
        return lastdir

    def save_current_index_dir(self, dir: str) -> None:
        """
        Saves the directory where we last indexed files.
        """
        self.settings.beginGroup("IndexDir")
        self.settings.setValue('dir', dir)
        self.settings.endGroup()

    def load_last_indexdir(self) -> str:
        self.settings.beginGroup('IndexDir')
        lastdir = self.settings.value("dir", type=str)
        if not Path(lastdir).exists():
            lastdir = './'
        self.settings.endGroup()
        return lastdir

    def save_ccdc_exe_path(self, path: str) -> None:
        """
        Saves the last directory where we saved a database.
        """
        self.settings.beginGroup("CCDC")
        self.settings.setValue('exepath', path)
        self.settings.endGroup()

    def load_ccdc_exe_path(self) -> str:
        self.settings.beginGroup('CCDC')
        exe_path = self.settings.value("exepath", type=str)
        if not Path(exe_path).exists() or not Path(exe_path).is_file():
            exe_path = ''
        self.settings.endGroup()
        return exe_path



from PyQt5 import QtWidgets, QtGui
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QCursor


class StructuresListTableView(QtWidgets.QTableView):
    save_excel_triggered = pyqtSignal()

    def __init__(self, parent):
        super(StructuresListTableView, self).__init__(parent)
        self.setContextMenuPolicy(Qt.DefaultContextMenu)

    def contextMenuEvent(self, event: QtGui.QContextMenuEvent) -> None:
        context_menu = QtWidgets.QMenu(self)
        save_excel = context_menu.addAction("Save as Excel File")
        save_excel.triggered.connect(lambda: self._on_save_excel(event))
        context_menu.addAction(save_excel)
        context_menu.popup(QCursor.pos())

    def _on_save_excel(self, event: QtGui.QContextMenuEvent):
        self.save_excel_triggered.emit()

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:
        if e.button() == Qt.RightButton:
            pass
        super(StructuresListTableView, self).mousePressEvent(e)



import datetime
import math
import os
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import List, Dict, Union, Tuple
from xml.etree.ElementTree import ParseError

from gemmi.cif import Style

from structurefinder.ccdc.query import get_cccsd_path, search_csd, parse_results
from structurefinder.misc.exporter import cif_data_to_document
from structurefinder.misc.version import VERSION
from structurefinder.searcher.constants import centering_letter_2_num, centering_num_2_letter
from structurefinder.searcher.database_handler import StructureTable
from structurefinder.searcher.misc import is_valid_cell, format_sum_formula, regular_results_parameters, vol_unitcell, \
    get_list_of_elements, more_results_parameters, is_a_nonzero_file, combine_results

parser = ArgumentParser(prog='strf_web',
                        description=f'StructureFinder Web Server v{VERSION}')
parser.add_argument('-n', '--host', default='127.0.0.1', type=str, dest='host')
parser.add_argument('-p', '--port', default='8080', type=str, dest='port')
parser.add_argument('-f', '--dbfile', default='structuredb.sqlite', dest='dbfilename')
parser.add_argument('-d', '--download', dest='download_button', action='store_true',
                    help='Shows a download link in the page bottom')
args = parser.parse_args()

host = args.host
port = args.port
# Give an absolute path:
dbfilename = Path(args.dbfilename).resolve()
download_button = args.download_button

###########################################################

site_ip = host + ':' + port
os.chdir(Path(__file__).parent.parent)
try:  # Adding local path to PATH
    sys.path.insert(0, os.path.abspath('.'))
except(KeyError, ValueError):
    print('Unable to set PATH properly. strf_web.py might not work.')

pyver = sys.version_info
if pyver[0] == 3 and pyver[1] < 4:
    # Python 2 creates a syntax error anyway.
    print("You need Python 3.4 and up in oder to run this program!")
    sys.exit()

from shutil import which

from structurefinder.cgi_ui.bottle import Bottle, static_file, template, redirect, request, response, HTTPResponse
from structurefinder.displaymol.mol_file_writer import MolFile
from structurefinder.displaymol.sdm import SDM
from structurefinder.pymatgen.core import lattice

app = application = Bottle()


@app.get('/all')
def structures_list_data():
    """
    The content of the structures list.
    """
    structures = StructureTable(dbfilename)
    return get_structures_json(structures, show_all=True)


@app.get('/')
def main():
    """
    The main web site with html template.
    """
    response.set_header('Set-Cookie', 'str_id=')
    response.content_type = 'text/html; charset=UTF-8'
    data = {"my_ip"        : site_ip,
            "title"        : 'StructureFinder',
            'host'         : host,
            'download_link': r"""<p><a href="http://{}/dbfile.sqlite" download="structurefinder.sqlite" 
                                     type="application/*">Download
                                    database file</a></p>""".format(site_ip) if download_button else ''
            }
    output = template('cgi_ui/views/strf_web', data)
    return output


@app.get('/dbfile.sqlite')
def get_dbfile():
    if download_button:
        return Path(dbfilename).read_bytes()
    else:
        return error404('File not found.')


@app.get("/cellsrch")
def cellsrch():
    cell_search = request.GET.cell_search
    more_results = (request.GET.more == "true")
    sublattice = (request.GET.supercell == "true")
    cell = is_valid_cell(cell_search)
    print("Cell search:", cell)
    structures = StructureTable(dbfilename)
    if cell:
        ids = find_cell(structures, cell, more_results=more_results, sublattice=sublattice)
        print("--> Got {} structures from cell search.".format(len(ids)))
        return get_structures_json(structures, ids)


@app.get("/txtsrch")
def txtsrch():
    structures = StructureTable(dbfilename)
    text_search = request.GET.text_search
    print("Text search:", text_search)
    ids = search_text(structures, text_search)
    return get_structures_json(structures, ids)


@app.get("/adv_srch")
def adv():
    elincl = request.GET.elements_in
    elexcl = request.GET.elements_out
    date1 = request.GET.date1
    date2 = request.GET.date2
    cell_search = request.GET.cell_search
    txt_in = request.GET.text_in
    txt_out = request.GET.text_out
    if len(txt_in) >= 2 and "*" not in txt_in:
        txt_in = '*' + txt_in + '*'
    if len(txt_out) >= 2 and "*" not in txt_out:
        txt_out = '*' + txt_out + '*'
    more_results = (request.GET.more == "true")
    sublattice = (request.GET.supercell == "true")
    onlyelem = (request.GET.onlyelem == "true")
    it_num = request.GET.it_num
    r1val = request.GET.r1val
    ccdc_num = request.GET.ccdc_num
    structures = StructureTable(dbfilename)
    print("Advanced search: elin:", elincl, 'elout:', elexcl, date1, '|', date2, '|', cell_search, 'txin:', txt_in,
          'txout:', txt_out, '|', 'more:', more_results, 'Sublatt:', sublattice, 'It-num:', it_num, 'only:', onlyelem,
          'CCDC:', ccdc_num)
    ids = advanced_search(cellstr=cell_search, elincl=elincl, elexcl=elexcl, txt=txt_in, txt_ex=txt_out,
                          sublattice=sublattice, more_results=more_results, date1=date1, date2=date2,
                          structures=structures, it_num=it_num, onlythese=onlyelem, r1val=r1val, ccdc_num=ccdc_num)
    print("--> Got {} structures from Advanced search.".format(len(ids)))
    return get_structures_json(structures, ids)


@app.post('/molecule')
def jsmol_request():
    """
    A request for atom data from jsmol.
    """
    str_id = request.POST.id
    print("Molecule id:", str_id)
    structures = StructureTable(dbfilename)
    if str_id:
        cell = structures.get_cell_by_id(str_id)
        if request.POST.grow == 'true':
            symmcards = [x.split(',') for x in structures.get_row_as_dict(str_id)
            ['_space_group_symop_operation_xyz'].replace("'", "").replace(" ", "").split("\n")]
            atoms = structures.get_atoms_table(str_id, cartesian=False, as_list=True)
            if atoms:
                sdm = SDM(atoms, symmcards, cell)
                needsymm = sdm.calc_sdm()
                atoms = sdm.packer(sdm, needsymm)
        else:
            atoms = structures.get_atoms_table(str_id, cartesian=True, as_list=False)
        try:
            m = MolFile(atoms)
            return m.make_mol()
        except(KeyError, TypeError) as e:
            print('Exception in jsmol_request: {}'.format(e))
            return ''


@app.post('/residuals')
def post_request():
    """
    Handle POST requests.
    """
    cif_dic = {}
    str_id = request.POST.id
    response.set_header('Set-Cookie', 'str_id=' + str_id)
    resid1 = request.POST.residuals1 == 'true'
    resid2 = request.POST.residuals2 == 'true'
    all_cif = request.POST.all == 'true'
    unitcell = request.POST.unitcell
    structures = StructureTable(dbfilename)
    print("Structure id:", str_id)
    if str_id:
        cif_dic = structures.get_row_as_dict(str_id)
    if str_id and unitcell and not (resid1 or resid2 or all_cif):
        try:
            return get_cell_parameters(structures, str_id)
        except ValueError as e:
            print("Exception raised:")
            print(e)
            return ''
    if str_id and resid1:
        return get_residuals_table1(structures, cif_dic, str_id)
    if str_id and resid2:
        return get_residuals_table2(cif_dic)
    if str_id and all_cif:
        return get_all_cif_val_table(structures, str_id)


# noinspection PyUnresolvedReferences
@app.route('/static/<filepath:path>')
def server_static(filepath):
    """
    Static files such as images or CSS files are not served automatically.
    The static_file() function is a helper to serve files in a safe and convenient way (see Static Files).
    This example is limited to files directly within the /path/to/your/static/files directory because the
    <filename> wildcard won’t match a path with a slash in it. To serve files in subdirectories, change
    the wildcard to use the path filter:
    """
    response = static_file(filepath, root='./cgi_ui/static/')
    response.set_header("Cache-Control", "public, max-age=240")
    return response


@app.route('/version')
def version():
    from structurefinder.misc.version import VERSION
    return 'version ' + str(VERSION)


@app.get('/cellcheck')
def cellsearch():
    if sys.platform == 'win32':
        if not get_cccsd_path():
            return 'false'
        else:
            return 'true'
    else:
        try:
            if which('ccdc_searcher') or \
                Path('/opt/CCDC/CellCheckCSD/bin/ccdc_searcher').exists():
                print('CellCheckCSD found')
                return 'true'
        except TypeError:
            return 'false'


@app.route('/favicon.ico')
def redirect_to_favicon():
    redirect('/static/favicon.ico')


@app.route('/current-cif/<structure_id:int>')
def download_currently_selected_cif(structure_id):
    if not download_button:
        return 'Downloading a CIF was turned off by the administrator.'
    headers = dict()
    structures = StructureTable(dbfilename)
    cif_data = structures.get_cif_export_data(structure_id)
    doc = cif_data_to_document(cif_data)
    file = doc.as_string(style=Style.Indent35)
    headers['Content-Type'] = 'text/plain'
    headers['Content-Encoding'] = 'ascii'
    headers['Content-Length'] = len(file)
    now = datetime.datetime.now()
    lm = now.strftime("%a, %d %b %Y %H:%M:%S GMT")
    headers['Last-Modified'] = lm
    return HTTPResponse(file, **headers)


@app.get('/csd')
def show_cellcheck():
    """
    Shows the CellcheckCSD web page
    """
    structures = StructureTable(dbfilename)
    str_id = request.get_cookie('str_id')
    centering = ''
    if str_id:
        cell = structures.get_cell_by_id(str_id)
        cif_dic = structures.get_row_as_dict(str_id)
        try:
            centering = cif_dic['_space_group_centring_type']
        except KeyError:
            centering = ''
        cellstr = '{:>8.3f} {:>8.3f} {:>8.3f} {:>8.3f} {:>8.3f} {:>8.3f}'.format(*cell)
    else:
        cellstr = ''
    if centering:
        try:
            cent = centering_letter_2_num[centering]
        except KeyError:  # mostly value of '?'
            cent = 0
    else:
        cent = 0
    response.content_type = 'text/html; charset=UTF-8'
    data = {"my_ip"  : site_ip,
            "title"  : 'StructureFinder',
            'cellstr': cellstr,
            'strid'  : str_id,
            'cent'   : cent,
            'host'   : host, }
    output = template('cgi_ui/views/cellcheckcsd', data)
    return output


@app.post('/csd-list')
def search_cellcheck_csd():
    """
    Search with CellcheckCSD.
    """
    cmd = request.POST.cmd
    cell = request.POST.cell
    str_id = request.POST.str_id
    if not cell:
        return {}
    cent = request.POST.centering
    if len(cell) < 6:
        return {}
    if cmd == 'get-records' and len(cell.split()) == 6:
        xml = search_csd(cell.split(), centering=centering_num_2_letter[int(cent)])
        # print(xml)
        try:
            results = parse_results(xml)  # results in a dictionary
        except ParseError as e:
            print(e)
            return
        # print(results)
        if str_id:
            structures = StructureTable(dbfilename)
        print(len(results), 'Structures found...')
        return {"total": len(results), "records": results, "status": "success"}
    else:
        return {}


@app.error(404)
def error404(error='Nothing here, sorry.'):
    """
    Redefine 404 message.
    """
    return f'''<div style="text-align: center;">
                <b>{error}</b><br>
                <p>
                <a href="http://{host}:{port}/">Back to main page</a>
                </p>
              </div>
            '''


def is_ajax():
    if request.headers.get('X-Requested-With') == 'XMLHttpRequest':
        return True
    else:
        return False


def get_structures_json(structures: StructureTable, ids: (list, tuple) = None, show_all: bool = False) -> dict:
    """
    Returns the next package of table rows for continuos scrolling.
    """
    if not ids and not show_all:
        return {}
    dic = structures.get_all_structures_as_dict(ids)
    number = len(dic)
    print("--> Got {} structures from actual search.".format(number))
    if number == 0:
        return {}
    return {"total": number, "records": dic, "status": "success"}


def get_cell_parameters(structures: StructureTable, strid: str) -> str:
    """
    Resturns unit cell parameters as html formated string.
    """
    c = structures.get_cell_by_id(strid)
    cstr = """<b>Unit Cell:</b>&nbsp;&nbsp; 
                      <i>a</i> = {0:>8.3f}&nbsp;&angst;,&nbsp;
                      <i>b</i> = {1:>8.3f}&nbsp;&angst;,&nbsp;
                      <i>c</i> = {2:>8.3f}&nbsp;&angst;,&nbsp; 
                      <i>&alpha;</i> = {3:>8.3f}&deg;,&nbsp;
                      <i>&beta;</i> = {4:>8.3f}&deg;,&nbsp;
                      <i>&gamma;</i> = {5:>8.3f}&deg;,&nbsp;
                      <i>V</i> = {6}&nbsp;&angst;<sup>3</sup>&nbsp;&nbsp;&nbsp;&nbsp; 
            <div style="font-size:0pt" id='hidden-cell'>{0}  {1}  {2}  {3}  {4}  {5}</div>
            """.format(c[0], c[1], c[2], c[3], c[4], c[5], round(c[6], 2))
    return cstr


def get_residuals_table1(structures: StructureTable, cif_dic: dict, structure_id: int) -> str:
    """
    Returns a table with the most important residuals of a structure.
    """
    try:
        rsigma = " / {}".format(cif_dic['_diffrn_reflns_av_unetI_netI'])
    except (TypeError, ValueError):
        rsigma = " "
    if not cif_dic:
        return ""
    if cif_dic['_refine_diff_density_max']:
        peakhole = "{} / {}".format(cif_dic['_refine_diff_density_max'], cif_dic['_refine_diff_density_min'])
    else:
        peakhole = " "
    try:
        sumform = format_sum_formula(structures.get_calc_sum_formula(structure_id), break_after=99)
    except KeyError:
        sumform = ''
    if sumform == '':
        # Display this as last resort:
        sumform = cif_dic['_chemical_formula_sum']
    table1 = """
    <table class="table table-bordered table-condensed" id='resitable1'>
        <tbody>
        <tr><td style='width: 40%'><b>Space Group</b></td>                 <td>{0}</td></tr>
        <tr><td><b>Z</b></td>                           <td>{1}</td></tr>
        <tr><td><b>Sum Formula</b></td>                 <td>{2}</td></tr>
        <tr><td><b>Temperature [K]</b></td>             <td>{3}</td></tr>
        <tr><td><b><i>wR</i><sub>2</sub></b></td>       <td>{4}</td></tr>
        <tr><td><b><i>R<i/><sub>1</sub></b></td>        <td>{5}</td></tr>
        <tr><td><b>Goof</b></td>                        <td>{6}</td></tr>
        <tr><td><b>Max Shift/esd</b></td>               <td>{7}</td></tr>
        <tr><td><b>Peak / Hole [e&angst;<sup>&minus;3</sup>]</b></td>             <td>{8}</td></tr>
        <tr><td><b><i>R</i><sub>int</sub> / <i>R</i><sub>&sigma;</sub></b></b></td>    <td>{9}{10} </td></tr>
        <tr><td><b>Wavelength [&angst;]</b></td>                      <td>{11}</td></tr>
        </tbody>
    </table>
    """.format(cif_dic['_space_group_name_H_M_alt'],
               cif_dic['_cell_formula_units_Z'],
               sumform,
               cif_dic['_diffrn_ambient_temperature'],
               cif_dic['_refine_ls_wR_factor_ref'] if cif_dic['_refine_ls_wR_factor_ref'] else cif_dic[
                   '_refine_ls_wR_factor_gt'],
               cif_dic['_refine_ls_R_factor_gt'] if cif_dic['_refine_ls_R_factor_gt'] else cif_dic[
                   '_refine_ls_R_factor_all'],
               cif_dic['_refine_ls_goodness_of_fit_ref'],
               cif_dic['_refine_ls_shift_su_max'],
               peakhole,
               cif_dic['_diffrn_reflns_av_R_equivalents'],
               rsigma,
               cif_dic['_diffrn_radiation_wavelength']
               )
    return table1


def get_residuals_table2(cif_dic: dict) -> str:
    """
    Returns a table with the most important residuals of a structure.
    """
    if not cif_dic:
        return ""
    wavelen = cif_dic['_diffrn_radiation_wavelength']
    thetamax = cif_dic['_diffrn_reflns_theta_max']
    thetafull = cif_dic['_diffrn_reflns_theta_full']
    # d = lambda/2sin(theta):
    try:
        d = wavelen / (2 * math.sin(math.radians(thetamax)))
    except(ZeroDivisionError, TypeError):
        d = 0.0
    try:
        compl = cif_dic['_diffrn_measured_fraction_theta_max'] * 100
        if not compl:
            compl = 0.0
        if isinstance(compl, str):
            compl = 0.0
    except TypeError:
        compl = 0.0
    try:
        data_to_param = cif_dic['_refine_ls_number_reflns'] / cif_dic['_refine_ls_number_parameters']
    except TypeError:
        data_to_param = 0
    table2 = """
    <table class="table table-bordered table-condensed" id='resitable2'>
        <tbody>
        <tr><td style='width: 40%'><b>Measured Refl.</b></td>       <td>{0}</td></tr>
        <tr><td><b>Independent Refl.</b></td>                       <td>{10}</td></tr>
        <tr><td><b>Data with [<i>I</i>>2&sigma;(<i>I</i>)] </b></td>    <td>{11}</td></tr>
        <tr><td><b>Parameters</b></td>                              <td>{1}</td></tr>
        <tr><td><b>data/param</b></td>                              <td>{2:<5.1f}</td></tr>
        <tr><td><b>Restraints</b></td>                              <td>{3}</td></tr>
        <tr><td><b>&theta;<sub>full</sub> [&deg;]</b> / 
        <b>&theta;<sub>max</sub> [&deg;]</b></td>                    <td>{4} / {5}</td></tr>
        <tr><td><b>d [&angst;]</b></td>                             <td>{6:5.3f}</td></tr>
        <tr><td><b>completeness [%]</b></td>                            <td>{7:<5.1f}</td></tr>
        <tr><td><b>Flack X parameter</b></td>                             <td>{8}</td></tr>
        <tr><td><b>CCDC Number</b></td>                             <td>{9}</td></tr>
        </tbody>
    </table>
    """.format(cif_dic['_diffrn_reflns_number'],  # 0
               cif_dic['_refine_ls_number_parameters'],  # 1
               data_to_param,  # 2
               cif_dic['_refine_ls_number_restraints'],  # 3
               thetamax if thetamax else '?',  # 4
               thetafull if thetafull else '?',  # 5
               d,  # 6
               compl,  # 7
               cif_dic['_refine_ls_abs_structure_Flack'],  # 8
               cif_dic['_database_code_depnum_ccdc_archive'],  # 9
               cif_dic['_refine_ls_number_reflns'],  # 10
               cif_dic['_reflns_number_gt']  # 11
               )
    return table2


def get_all_cif_val_table(structures: StructureTable, structure_id: int) -> str:
    """
    Returns a html table with the residuals values of a structure.
    """
    # starting table header (the div is for css):
    # style="white-space: pre": preserves white space
    button = """<a type="button" class="btn btn-default btn-sm" id="download_CIF"
                                                 href='current-cif/{}' >Download as CIF</a>""".format(structure_id)
    table_string = """<h4>All CIF values</h4> {}
                        <div id="myresidualtable">
                        <table class="table table-striped table-bordered table-condensed" style="white-space: pre">
                            <thead>
                                <tr>
                                    <th> Item </th>
                                    <th> Value </th>
                                </tr>
                            </thead>
                        <tbody>""".format(button if download_button else '')
    # get the residuals of the cif file as a dictionary:
    dic = structures.get_row_as_dict(structure_id)
    if not dic:
        return ""
    # filling table with data rows:
    for key, value in dic.items():
        if key == "Id":
            continue
        if isinstance(value, str):
            value = ''.join([x.replace("\n", "<br>").rstrip('\r\n') for x in value])
        if key == '_shelx_res_file':
            # Adding an ID to make font monospace in the view:
            table_string += '''<tr>
                                 <td class="residual-{}"> {} </a></td> 
                                 <td id=resfile > {} </a></td> 
                               </tr> \n'''.format(structure_id, key, value)
        else:
            table_string += '''<tr>
                                <td class="residual-{}"> {} </a></td> 
                                <td> {} </a></td> 
                           </tr> \n'''.format(structure_id, key, value)
    # closing table:
    table_string += """ </tbody>
                        </table>
                        </div>"""
    return table_string


def chunks(l: list, n: int) -> list:
    """
    returns successive n-sized chunks from l.
    >>> l = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 'a', 'b', 'c', 'd', 'e', 'f']
    >>> chunks(l, 5)
    [[1, 2, 3, 4, 5], [6, 7, 8, 9, 0], ['a', 'b', 'c', 'd', 'e'], ['f']]
    >>> chunks(l, 1)
    [[1], [2], [3], [4], [5], [6], [7], [8], [9], [0], ['a'], ['b'], ['c'], ['d'], ['e'], ['f']]
    >>> chunks(l, 50)
    [[1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 'a', 'b', 'c', 'd', 'e', 'f']]
    """
    return [l[i:i + n] for i in range(0, len(l), n)]


def find_cell(structures: StructureTable, cell: list, sublattice=False, more_results=False) -> list:
    """
    Finds unit cells in db. Rsturns hits a a list of ids.
    """
    volume = vol_unitcell(*cell)
    if more_results:
        # more results:
        atol, ltol, vol_threshold = more_results_parameters(volume)
    else:
        # regular:
        atol, ltol, vol_threshold = regular_results_parameters(volume)
    cells: List = structures.find_by_volume(volume, vol_threshold)
    if sublattice:
        # sub- and superlattices:
        for v in [volume * x for x in [2.0, 3.0, 4.0, 6.0, 8.0, 10.0]]:
            # First a list of structures where the volume is similar:
            cells.extend(structures.find_by_volume(v, vol_threshold))
        cells = list(set(cells))
    idlist2: List = []
    # Real lattice comparing in G6:
    if cells:
        lattice1 = lattice.Lattice.from_parameters(*cell)
        for num, curr_cell in enumerate(cells):
            try:
                lattice2 = lattice.Lattice.from_parameters(*curr_cell[1:7])
            except ValueError:
                continue
            mapping = lattice1.find_mapping(lattice2, ltol, atol, skip_rotation_matrix=True)
            if mapping:
                idlist2.append(curr_cell[0])
    if idlist2:
        return idlist2
    else:
        return []


def search_text(structures: StructureTable, search_string: str) -> tuple:
    """
    searches db for given text
    """
    idlist = []
    if len(search_string) == 0:
        return ()
    if len(search_string) >= 2 and "*" not in search_string:
        search_string = "{}{}{}".format('*', search_string, '*')
    try:
        #  bad hack, should make this return ids like cell search
        idlist = structures.find_text_and_authors(search_string)
    except AttributeError as e:
        print("Exception in search_text:")
        print(e)
    return idlist


def search_elements(structures: StructureTable, elements: str, excluding: str = '', onlyelem: bool = False) -> list:
    """
    list(set(l).intersection(l2))
    """
    res = []
    try:
        formula = get_list_of_elements(elements)
    except KeyError:
        print('Element search error! Wrong list of elements.')
        return []
    try:
        formula_ex = get_list_of_elements(excluding)
    except KeyError:
        print('Error: Wrong list of Elements!')
        return []
    try:
        res = structures.find_by_elements(formula, excluding=formula_ex, onlyincluded=onlyelem)
    except AttributeError:
        print('Element search error! Wrong list of elements..')
    return list(res)


def find_dates(structures: StructureTable, date1: str, date2: str) -> list:
    """
    Returns a list if id between date1 and date2
    """
    if not date1:
        date1 = '0000-01-01'
    if not date2:
        date2 = 'NOW'
    result = structures.find_by_date(date1, date2)
    return result


def advanced_search(cellstr: str, elincl, elexcl, txt, txt_ex, sublattice, more_results,
                    date1: str = None, date2: str = None, structures: StructureTable = None,
                    it_num: str = None, onlythese: bool = False, r1val: float = 0.0, ccdc_num: str = '') -> list:
    """
    Combines all the search fields. Collects all includes, all excludes ad calculates
    the difference.
    """
    #
    results: List = []
    cell_results: List = []
    spgr_results: List = []
    elincl_results: List = []
    txt_results: Union[List, Tuple] = []
    txt_ex_results: Union[List, Tuple] = []
    date_results: List = []
    ccdc_num_results: List = []
    states: Dict[str, bool] = {'date'    : False,
                               'cell'    : False,
                               'elincl'  : False,
                               'elexcl'  : False,
                               'txt'     : False,
                               'txt_ex'  : False,
                               'spgr'    : False,
                               'rval'    : False,
                               'ccdc_num': False,
                               }
    if ccdc_num:
        ccdc_num_results = structures.find_by_ccdc_num(ccdc_num)
    if ccdc_num_results:
        return ccdc_num_results
    cell = is_valid_cell(cellstr)
    try:
        spgr = int(it_num.split()[0])
    except Exception:
        spgr = 0
    try:
        rval = float(r1val)
        states['rval'] = True
    except ValueError:
        rval = 0.0
    if cell:
        states['cell'] = True
        cell_results = find_cell(structures, cell, sublattice=sublattice, more_results=more_results)
    if spgr:
        states['spgr'] = True
        spgr_results = structures.find_by_it_number(spgr)
    if elincl or elexcl:
        if elincl:
            states['elincl'] = True
        if elexcl:
            states['elexcl'] = True
        elincl_results = search_elements(structures, elincl, elexcl, onlythese)
    if txt:
        states['txt'] = True
        txt_results = structures.find_text_and_authors(txt)
    if txt_ex:
        states['txt_ex'] = True
        txt_ex_results = structures.find_text_and_authors(txt_ex)
    if date1 != date2:
        states['date'] = True
        date_results = find_dates(structures, date1, date2)
    rval_results = []
    if rval > 0.0:
        rval_results = structures.find_by_rvalue(rval / 100)
    ####################
    results = combine_results(cell_results=cell_results,
                              date_results=date_results,
                              elincl_results=elincl_results,
                              results=results,
                              spgr_results=spgr_results,
                              txt_ex_results=txt_ex_results,
                              txt_results=txt_results,
                              rval_results=rval_results,
                              states=states)
    return list(results)


def run():
    print(f"----- StructureFinder web application version {sys.version} --------------")
    print(f" Running on Python version {sys.version}")
    if not is_a_nonzero_file(dbfilename):
        print("Unable to start!")
        print(f"The database file '{os.path.abspath(dbfilename)}' does not exist.")
        sys.exit()
    print(f'### Running with database "{os.path.abspath(dbfilename)}" ###')
    # plain python wsgiref server (gunicorn doesnt run on windows):
    # app.run(host=host, port=port, server='wsgiref', reloader=True)
    # gunicorn server: Best used behind an nginx proxy server: http://docs.gunicorn.org/en/stable/deploy.html
    # you need "pip3 install gunicorn" to run this:
    # The current database interface allows only one worker (have to go to sqlalchemy!)
    if sys.platform == 'win32':
        server = 'wsgiref'
    else:
        server = 'gunicorn'
    app.run(host=host, port=port, reload=True, server=server, accesslog='-', errorlog='-', workers=1,
            access_log_format='%(h)s %(l)s %(u)s %(t)s "%(r)s" %(s)s %(b)s "%(f)s"')


if __name__ == "__main__":
    run()



import ctypes
import os
from pathlib import Path

from PyQt5.QtWidgets import QMainWindow, QMessageBox

from structurefinder.misc.version import VERSION


def bug_found_warning(logfile) -> None:
    window = QMainWindow()
    title = f'Congratulations, you found a bug in StructureFinder!'
    text = (f'<br>Please send the file <br><br>'
            f'<a href=file:{os.sep * 2}{logfile.resolve()}>{logfile.resolve()}</a> '
            f'<br><br>to Daniel Kratzert: '
            f'<a href="mailto:dkratzert@gmx.de?subject=StructureFinder version {VERSION} crash report">'
            f'dkratzert@gmx.de</a><br>')
    box = QMessageBox(parent=window)
    box.setWindowTitle('Warning')
    box.setText(title)
    box.setInformativeText(text)
    box.exec()
    window.show()


def do_update_program(version: str):
    # parent path of gui -> main dir:
    updater_exe = str(Path(__file__).parent.parent.parent.joinpath('update.exe'))
    args = ['-v', version,
            '-p', 'structurefinder']
    # Using this, because otherwise I can not write to the program dir:
    ctypes.windll.shell32.ShellExecuteW(None, "runas", updater_exe, " ".join(args), None, 1)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Created on 09.02.2015

 ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE" (Revision 42):
* <dkratzert@gmx.de> wrote this file. As long as you retain this
* notice you can do whatever you want with this stuff. If we meet some day, and
* you think this stuff is worth it, you can buy me a beer in return.
* ----------------------------------------------------------------------------

@author: Daniel Kratzert
"""

import fnmatch
import os
import re
import tarfile
import zipfile
from typing import Generator, Tuple

import gemmi

from structurefinder.searcher import database_handler
from structurefinder.searcher.fileparser import CifFile
from structurefinder.searcher.misc import get_value
from structurefinder.shelxfile.shelx import ShelXFile

DEBUG = False

excluded_names = ['ROOT',
                  '.OLEX',
                  'olex',
                  'TMP',
                  'TEMP',
                  'Papierkorb',
                  'Recycle.Bin',
                  'dsrsaves',
                  'BrukerShelXlesaves',
                  'shelXlesaves'
                  ]


class MyZipBase(object):
    def __init__(self, filepath: str) -> None:
        self.filepath = filepath
        self.cifname = ''
        self.cifpath = ''


class MyZipReader(MyZipBase):
    def __init__(self, filepath):
        """
        extracts .cif files from zip files
        """
        super().__init__(filepath)

    def __iter__(self) -> Generator:
        """
        returns an iterator of cif files in the zipfile as list.
        """
        try:
            zfile = zipfile.ZipFile(self.filepath)
            for name in zfile.namelist():
                (self.cifpath, self.cifname) = os.path.split(name)
                if self.cifname.endswith('.cif'):
                    if not self.cifname.startswith('__') and zfile.NameToInfo[name].file_size < 150000000:
                        doc = gemmi.cif.Document()
                        doc.source = name
                        doc.parse_string(zfile.read(name).decode('ascii', 'ignore'))
                        if doc:
                            yield doc
                        else:
                            continue
        except Exception as e:
            # print("Error: '{}' in file {}".format(e, self.filepath.encode(encoding='utf-8', errors='ignore')))
            # print(e, self.filepath)  # filepath is not utf-8 save
            yield None


class MyTarReader(MyZipBase):
    def __init__(self, filepath):
        """
        extracts .cif files from tar.gz files
        """
        super().__init__(filepath)

    def __iter__(self) -> Generator:
        """
        returns an iterator of cif files in the zipfile as list.
        """
        try:
            tfile = tarfile.open(self.filepath, mode='r')
            for name in tfile.getnames():
                self.cifpath, self.cifname = os.path.split(name)
                if self.cifname.endswith('.cif'):
                    doc = gemmi.cif.Document()
                    doc.source = name
                    doc.parse_string(tfile.extractfile(name).read().decode('ascii', 'ignore'))
                    if doc:
                        yield doc
                    else:
                        continue
        except Exception as e:
            # print("Error: '{}' in file {}".format(e, self.filepath.encode(encoding='utf-8', errors='ignore')))
            # print(e, self.filepath)  # filepath is not utf-8 save
            yield None


def filewalker_walk(startdir: str, patterns: list) -> Tuple[Tuple[str, str]]:
    """
    walks through the filesystem starting from startdir and searches
    for files with ending endings.

    Since os.walk() uses scandir, it is as fast as pathlib.
    """
    filelist = []
    print('collecting {} files below '.format(', '.join(patterns)) + startdir)
    for root, _, files in os.walk(startdir):
        for filen in files:
            omit = False
            if any(fnmatch.fnmatch(filen, pattern) for pattern in patterns):
                for ex in excluded_names:
                    if re.search(ex, root, re.I):
                        omit = True
                        break
                if omit:
                    continue
                fullpath = os.path.abspath(os.path.join(root, filen))
                try:
                    if os.stat(fullpath).st_size == 0:
                        continue
                except Exception:
                    continue
                if filen == 'xd_geo.cif':  # Exclude xdgeom cif files
                    continue
                if filen == 'xd_four.cif':  # Exclude xdfourier cif files
                    continue
                # This is much faster than yield():
                filelist.append((root, filen))
            else:
                continue
    return tuple(filelist)


def fill_db_with_cif_data(cif: CifFile, filename: str, path: str, structure_id: int,
                          structures: database_handler.StructureTable):
    """
    Fill all info from cif file into the database tables
    _atom_site_label
    _atom_site_type_symbol
    _atom_site_fract_x
    _atom_site_fract_y
    _atom_site_fract_z
    _atom_site_U_iso_or_equiv
    _atom_site_adp_type
    _atom_site_occupancy
    _atom_site_site_symmetry_order
    _atom_site_calc_flag
    _atom_site_refinement_flags_posn
    _atom_site_refinement_flags_adp
    _atom_site_refinement_flags_occupancy
    _atom_site_disorder_assembly
    _atom_site_disorder_group
    """
    sum_formula_dict = {}
    measurement_id = 1
    structures.fill_structures_table(path, filename, structure_id, measurement_id, cif.block.name)
    structures.fill_cell_table(structure_id, cif.cell.a, cif.cell.b, cif.cell.c,
                               cif.cell.alpha, cif.cell.beta, cif.cell.gamma, cif.cell.volume)
    try:
        sum_formula_dict = add_atoms(cif, structure_id, structures)
    except AttributeError as e:
        print('Atoms crashed', e, structure_id)
    cif.cif_data['calculated_formula_sum'] = sum_formula_dict
    structures.fill_residuals_table(structure_id, cif)
    structures.fill_authors_table(structure_id, cif)
    return True


def add_atoms(cif, structure_id, structures):
    sum_formula_dict = {}
    for at, orth in zip(cif.atoms, cif.atoms_orth):
        try:
            try:
                part = at.part
                if part in {'.', '', '?'}:
                    part = 0
            except (KeyError, ValueError, IndexError):
                part = 0
            try:
                occu = get_value(at.occ)
                if not occu:
                    occu = 1.0
            except (KeyError, ValueError, IndexError):
                occu = 1.0
            try:
                structures.fill_atoms_table(structure_id, at.label, at.type,
                                            get_value(at.x), get_value(at.y), get_value(at.z),
                                            occu, part,
                                            orth.x, orth.y, orth.z)
            except ValueError:
                pass
                # print(cif.cif_data['data'], structure_id)
            if at.type in sum_formula_dict:
                sum_formula_dict[at.type] += occu
            else:
                sum_formula_dict[at.type] = occu
        except KeyError as e:
            # print(at, structure_id, e)
            pass
    return sum_formula_dict


def fill_db_with_res_data(res: ShelXFile, filename: str, path: str, structure_id: int,
                          structures: database_handler.StructureTable, options: dict):
    if not res.cell:
        return False
    if not all([res.cell.a, res.cell.b, res.cell.c, res.cell.al, res.cell.be, res.cell.ga]):
        return False
    if not res.cell.volume:
        return False
    # Unused value:
    measurement_id = 1
    structures.fill_structures_table(path, filename, structure_id, measurement_id, res.titl)
    structures.fill_cell_table(structure_id, res.cell.a, res.cell.b, res.cell.c, res.cell.al,
                               res.cell.be, res.cell.ga, res.cell.volume)
    for at in res.atoms:
        if at.qpeak:
            continue
        if at.element.lower() == 'cnt':  # Do not add Shelxle centroids
            continue
        structures.fill_atoms_table(structure_id,
                                    at.name,
                                    at.element.capitalize(),
                                    at.x,
                                    at.y,
                                    at.z,
                                    at.sof,
                                    at.part.n,
                                    round(at.xc, 5), round(at.yc, 5), round(at.zc, 5))
    cif = CifFile(options=options)
    cif.cif_data["_cell_formula_units_Z"] = res.Z
    try:
        symmops = "\n".join([x.to_fractional() for x in res.symmcards])
        # gemmi needs integer ratio instead of float values in symmcards:
        # symmops = "\n".join([x.toShelxl() for x in res.symmcards])
        cif.cif_data["_space_group_symop_operation_xyz"] = symmops
    except Exception:
        try:
            symmops = "\n".join([x.toShelxl() for x in res.symmcards])
            cif.cif_data["_space_group_symop_operation_xyz"] = symmops
        except Exception:
            pass
    try:
        cif.cif_data["calculated_formula_sum"] = res.sum_formula_ex_dict()
    except ZeroDivisionError:
        pass
    try:
        cif.cif_data["_chemical_formula_sum"] = res.sum_formula_exact
    except ZeroDivisionError:
        pass
    cif.cif_data["_diffrn_radiation_wavelength"] = res.wavelen
    if res.R1:
        cif.cif_data["_refine_ls_R_factor_gt"] = res.R1
    if res.wR2:
        cif.cif_data["_refine_ls_wR_factor_ref"] = res.wR2
    if res.parameters:
        cif.cif_data['_refine_ls_number_parameters'] = res.parameters
    if res.data:
        cif.cif_data['_refine_ls_number_reflns'] = res.data
    if res.num_restraints:
        cif.cif_data['_refine_ls_number_restraints'] = res.num_restraints
    if res.temp_in_Kelvin:
        cif.cif_data['_diffrn_ambient_temperature'] = round(res.temp_in_Kelvin, 5)
    if res.dhole:
        cif.cif_data['_refine_diff_density_min'] = res.dhole
    if res.hpeak:
        cif.cif_data['_refine_diff_density_max'] = res.hpeak
    if res.latt:
        cif.cif_data['_space_group_centring_type'] = res.latt.N_str
    if cif.cif_data["_space_group_symop_operation_xyz"]:
        try:
            symm_ops = cif.cif_data["_space_group_symop_operation_xyz"].splitlines(keepends=False)
            spgr: gemmi.SpaceGroup = gemmi.find_spacegroup_by_ops(gemmi.GroupOps([gemmi.Op(o) for o in symm_ops]))
            if spgr:
                cif.cif_data["_space_group_name_H-M_alt"] = spgr.short_name()
                cif.cif_data['_space_group_IT_number'] = spgr.number
                cif.cif_data["_space_group_centring_type"] = spgr.centring_type()
        except RuntimeError as e:
            # print(e, filename)
            pass
    if res.space_group and not cif.cif_data["_space_group_name_H-M_alt"]:
        cif.cif_data["_space_group_name_H-M_alt"] = res.space_group
        try:
            spgr = gemmi.find_spacegroup_by_name(str(res.space_group).replace(')', '').replace('(', ''))
            if not cif.cif_data['_space_group_IT_number']:
                cif.cif_data['_space_group_IT_number'] = spgr.number
        except AttributeError:
            pass
    if res.goof:
        cif.cif_data["_refine_ls_goodness_of_fit_ref"] = res.goof
    if res.rgoof:
        cif.cif_data["_refine_ls_restrained_S_all"] = res.rgoof
    try:
        cif.cif_data["_shelx_res_file"] = str(res)
    except IndexError:
        pass
    structures.fill_residuals_table(structure_id, cif)
    return True


if __name__ == '__main__':
    z = MyTarReader('./tests/test-data/106c.tar.bz2')

    for i in z:
        print(i)

    # filewalker_walk('./')
    # z = zipopener('./tests/test-data/Archiv.zip')
    # print(z)

    # fp = create_file_list('../test-data/', 'zip')
    # for i in fp:
    #    print(i)



"""
MOl V3000 format
"""

import os

from structurefinder.searcher.misc import distance
from structurefinder.shelxfile.elements import get_radius_from_element


class MolFile(object):
    """
    This mol file writer is only to use the file with JSmol, not to implement the standard exactly!
    """

    def __init__(self, atoms: list, bonds = None):
        self.atoms = atoms
        if bonds:
            self.bonds = bonds
        else:
            self.bonds = self.get_conntable_from_atoms()
        self.bondscount = len(self.bonds)
        self.atomscount = len(self.atoms)

    def header(self) -> str:
        """
        For JSmol, I don't need a facy header.
        """
        return "{}{}{}".format(os.linesep, os.linesep, os.linesep)

    def connection_table(self) -> str:
        """
          6  6  0  0  0  0  0  0  0  0  1 V3000
        """
        tab = "{:>5d}{:>5d}".format(self.atomscount, self.bondscount)
        return tab

    def get_atoms_string(self) -> str:
        """
        Returns a string with an atom in each line.
        X Y Z Element
        """
        atoms = []
        for at in self.atoms:
            atoms.append("{:>10.4f}{:>10.4f}{:>10.4f} {:<2s}".format(at[2], at[3], at[4], at[1]))
        return '\n'.join(atoms)

    def get_bonds_string(self) -> str:
        """
        This is not accodingly to the file standard!
        The standard wants to have fixed format 3 digits for the bonds.
        """
        blist = []
        for bo in self.bonds:
            # This is deviating from the standard:
            blist.append("{:>4d}{:>4d}  1  0  0  0  0".format(bo[0], bo[1]))
        return '\n'.join(blist)

    def get_conntable_from_atoms(self, extra_param=0.48):
        """
        returns a connectivity table from the atomic coordinates and the covalence
        radii of the atoms.
        a bond is defined with less than the sum of the covalence radii plus the extra_param:
        :param extra_param: additional distance to the covalence radius
        :type extra_param: float
        """
        #t1 = perf_counter()
        conlist = []
        for num1, at1 in enumerate(self.atoms, 1):
            at1_part = at1[5]
            rad1 = get_radius_from_element(at1[1])
            for num2, at2 in enumerate(self.atoms, 1):
                at2_part = at2[5]
                if at1_part * at2_part != 0 and at1_part != at2_part:
                    continue
                if at1[0] == at2[0]:  # name1 = name2
                    continue
                d = distance(at1[2], at1[3], at1[4], at2[2], at2[3], at2[4])
                if d > 4.0:  # makes bonding faster (longer bonds do not exist)
                    continue
                rad2 = get_radius_from_element(at2[1])
                if (rad1 + rad2) + extra_param > d:
                    if at1[1] == 'H' and at2[1] == 'H':
                        continue
                    # print(num1, num2, d)
                    # The extra time for this is not too much:
                    if [num2, num1] in conlist:
                        continue
                    conlist.append([num1, num2])
        #t2 = perf_counter()
        #print('Bondzeit:', round(t2-t1, 3), 's')
        #print('len:', len(conlist))
        return conlist

    def footer(self) -> str:
        """
        """
        return "M  END{}$$$$".format(os.linesep)

    def make_mol(self):
        """
        Combines all above to a mol file.
        """
        header = '\n\n'
        connection_table = self.connection_table()
        atoms = self.get_atoms_string()
        bonds = self.get_bonds_string()
        footer = self.footer()
        mol = "{0}{5}{1}{5}{2}{5}{3}{5}{4}".format(header, connection_table, atoms, bonds, footer, '\n')
        return mol




# TODO:
# Calculate real space group from symm cards during indexing
# use following cell comparison
# store niggli cell during indexing

import gemmi
cell = gemmi.UnitCell(13.738, 14.735, 16.598, 116.361, 97.813, 90.693)
sg = gemmi.SpaceGroup('P 1')
gv = gemmi.GruberVector(cell, sg)

cell2 = gemmi.UnitCell(10.360, 18.037, 25.764, 127.030, 129.810, 90.510)
sg2 = gemmi.SpaceGroup('P 1')
gv2 = gemmi.GruberVector(cell2, sg2)

gv.niggli_reduce()
gv2.niggli_reduce()

gv.get_cell()
# <gemmi.UnitCell(13.738, 14.735, 16.5958, 63.6541, 81.5645, 89.307)>
gv2.get_cell()
# <gemmi.UnitCell(10.36, 12.9807, 18.037, 100.37, 90.51, 108.246)>

gvc = gv.get_cell()
gvc.approx(gv2.get_cell(), 0.0001)
# False

gvc.is_similar(gv2.get_cell(), rel=0.1, deg=2.5)
# True

sv = gemmi.SellingVector(cell, sg)
sv2 = gemmi.SellingVector(cell2, sg2)

sv2.reduce()
# 3
sv.reduce()
# 1
sv.sort()
sv2.sort()
print(sv)
# <gemmi.SellingVector((-108.60, -31.00, -2.45, -155.29, -106.08, -135.90))>
print(sv2)
# <gemmi.SellingVector((-42.15, -1.66, -42.11, -63.56, -84.25, -281.52))>

sv.sum_b_squared()
# 1078.609162185184
sv2.sum_b_squared()
# 1030.492643465581



from typing import Any

from PyQt5 import QtCore
from PyQt5.QtCore import QModelIndex, Qt


class TableModel(QtCore.QAbstractTableModel):
    def __init__(self, *args, structures=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.horizontalHeaders = ['Id', 'Data Name', 'File Name', 'Last Modified', 'Path']
        self._data = structures or []

    def data(self, index: QModelIndex, role: int = None):
        row, col = index.row(), index.column()
        value = self._data[row][col]
        if role == Qt.DisplayRole:
            if isinstance(value, bytes):
                return value.decode('utf-8')
            else:
                return value

    def setHeaderData(self, section, orientation, data, role=Qt.EditRole):
        if orientation == Qt.Horizontal and role in (Qt.DisplayRole, Qt.EditRole):
            try:
                self.horizontalHeaders[section] = data
                return True
            except IndexError:
                return False
        return super().setHeaderData(section, orientation, data, role)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            try:
                return self.horizontalHeaders[section]
            except IndexError:
                pass
        return super().headerData(section, orientation, role)

    def rowCount(self, parent=None, *args, **kwargs):
        """
        The length of the outer list.
        """
        return len(self._data)

    def columnCount(self, parent=None, *args, **kwargs):
        """
        Takes the first sub-list, and returns
        the length (only works if all rows are an equal length)
        """
        if len(self._data) > 0:
            return len(self._data[0])
        else:
            return 0

    def setData(self, index: QModelIndex, value: Any, role: int = None) -> bool:
        row, col = index.row(), index.column()
        if not index:
            return False
        if index.isValid() and role == Qt.EditRole:
            self._data[row][col] = value
            return True
        return False

    def clear(self):
        self.resetInternalData()

    def sort(self, column: int, order: Qt.SortOrder = ...) -> None:
        self.layoutAboutToBeChanged.emit()
        self._data.sort(key=lambda x: x[column], reverse=True if order == Qt.DescendingOrder else False)
        self.layoutChanged.emit()
        # super(TableModel, self).sort(column, order)



#!python3
"""
A spinner from https://stackoverflow.com/questions/4995733/how-to-create-a-spinning-command-line-cursor-using-python
"""

import sys
import time
import threading


class Spinner:
    busy = False
    delay = 0.1

    @staticmethod
    def spinning_cursor():
        while 1:
            for cursor in '|/-\\': yield cursor

    def __init__(self, delay=None):
        self.spinner_generator = self.spinning_cursor()
        if delay and float(delay): self.delay = delay

    def spinner_task(self):
        while self.busy:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write('\b')
            sys.stdout.flush()

    def start(self):
        self.busy = True
        threading.Thread(target=self.spinner_task).start()

    def stop(self):
        self.busy = False
        time.sleep(self.delay)




def write(moldata: str, width: int = 240, height: int = 240):
    m = """
    <!DOCTYPE html>
    <html>
    <head>
    <meta charset="utf-8">
    <script src="jquery.min.js"></script>
    <script src="JSmol_dk.nojq.lite.js"></script>
    
    <!-- 
    A template to display molecules with JSmol Lite.
    --> 
    """ + r"""
    <!-- MOL DATA
    {}
    -->
    """.format(moldata) + r"""
    
    <script type="text/javascript"> 
    
    
    $(window).ready(function() {
      var Info;
      Info = {
      """ + """
          width: {},
          height: {},
          """.format(width, height) + r"""
          color: "0xFFFFFF",
          shadeAtoms: false,
          addSelectionOptions: false,
          use: "HTML5",
          readyFunction: null,
          defaultModel: "",
          bondWidth: 2,
          zoomScaling: 8,
          pinchScaling: 6.0,
          mouseDragFactor: 0.7,
          touchDragFactor: 0.15,
          multipleBondSpacing: 0,
          spinRateX: -0.08,
          spinRateY: 0.05,
          spinFPS: 20,
          spin: false,
          debug: false,
          infodiv: false,
          j2sPath: "."
      };
      Jmol._document = null;
      Jmol.getTMApplet("jmol", Info);
      $("#apphere").html(jmol._code);
      var mol = $('html').html().split("MOL DATA\n")[1].split("-->")[0];
      jmol.__loadModel(mol);
    });

    </script>
    </head>
    <body>

    <div id="apphere"></div>

    </body>
    </html>
    """
    return m



# -*- encoding: utf-8 -*-
# möpß
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42):
# <dkratzert@gmx.de> wrote this file. As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.
# Daniel Kratzert
# ----------------------------------------------------------------------------
#

"""
This file reads Bruker p4p files into a data structure.
"""

from structurefinder.shelxfile.dsrmath import Matrix


def read_file_to_list(p4pfile: str) -> list:
    """
    Reads a file and returns a list without line endings.
    """
    p4plist = []
    try:
        with open(p4pfile, 'r') as f:
            p4plist = f.read().splitlines(keepends=False)
    except IOError as e:
        print(e)
        print('*** CANNOT READ FILE {} ***'.format(p4pfile))
    return p4plist


class P4PFile():

    def __init__(self, p4plist):
        self.p4plist = p4plist
        self.fileid = None
        self.siteid = None
        self.chem = None
        self.cell = None
        self.cellsd = None
        self.ort1 = None
        self.ort2 = None
        self.ort3 = None
        self.zeros = None
        self.source = None
        self.volume = None
        self.ortmatrix = None 
        try:
            self.parse_p4p()
        except Exception:
            print('*** p4p not readable ***')

    def parse_p4p(self):
        for line in self.p4plist:
            spline = line.split()
            card = spline[0]
            if card == "CELL" and len(spline) > 5:
                self.cell = [float(x) for x in spline[1:7]]
                if len(spline) > 6:
                    self.volume = float(spline[7])
            if card == "FILEID":
                self.fileid = spline[1:]
            if card == 'CELLSD':
                self.cellsd = self.to_float_list(spline[1:])
            if card == "ORT1":
                self.ort1 = self.to_float_list(spline[1:])
            if card == "ORT2":
                self.ort2 = self.to_float_list(spline[1:])
            if card == "ORT3":
                self.ort3 = self.to_float_list(spline[1:])
            if all([self.ort1, self.ort2, self.ort3]):
                self.ortmatrix = Matrix([self.ort1, self.ort2, self.ort3])
            if card == "CHEM":
                self.chem = spline[1]

    @staticmethod
    def to_float_list(items):
        return [float(x) for x in items]


if __name__ == '__main__':
    p4p = P4PFile('./tests/test-data/test1.p4p')
    print(p4p.cell)
    print(p4p.cellsd)
    print(p4p.volume)
    print(p4p.ortmatrix)
    print(p4p.chem)



import sys

import requests
from PyQt5.QtCore import QThread, pyqtSignal

from structurefinder.misc.version import VERSION


class MyDownloader(QThread):
    progress = pyqtSignal(str)
    failed = pyqtSignal(int)
    finished = pyqtSignal(bytes)

    def __init__(self, parent, url: str):
        super().__init__(parent)
        self.url = url

    def run(self):
        try:
            self.download(self.url)
        except requests.RequestException as e:
            print('Could not connect to download server')
            print(e)

    def print_status(self, status: str) -> None:
        #print(status)
        pass

    def failed_to_download(self, status_code: int):
        print('Failed to download: {}'.format(self.url))
        print('HTTP status was {}'.format(status_code))

    def download(self, full_url: str) -> bytes:
        OS = sys.platform
        user_agent = f'StructureFinder v{VERSION} ({OS})'
        headers = {
            'User-Agent': user_agent,
        }
        # noinspection PyUnresolvedReferences
        self.progress.emit('Starting download: {}'.format(full_url))
        response = requests.get(full_url, stream=True, headers=headers, timeout=5)
        if response.status_code != 200:
            # noinspection PyUnresolvedReferences
            self.failed.emit(response.status_code)
            # noinspection PyUnresolvedReferences
            self.finished.emit(b'')
            return b''
        # noinspection PyUnresolvedReferences
        self.finished.emit(response.content)
        return response.content


if __name__ == "__main__":
    from PyQt5.QtWidgets import QWidget
    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)
    w = QWidget()
    w.show()


    def foo(bar: bytes):
        print(bar.decode('ascii'))


    upd = MyDownloader(None, "https://dkratzert.de/files/structurefinder/version.txt")
    upd.finished.connect(foo)
    upd.failed.connect(upd.failed_to_download)
    upd.progress.connect(upd.print_status)
    upd.start()
    sys.exit(app.exec_())
